from .spectra import *
from .emission_regions import *
from .targets import *
from .synchrotron import *
from .compton import *
from .absorption import *
from .constraints import *
from .utils import *



# module containing the External Compton radiative process
import numpy as np
from astropy.constants import c, sigma_T, G
from ..utils.math import (
    axes_reshaper,
    gamma_to_integrate,
    mu_to_integrate,
    phi_to_integrate,
)
from ..utils.conversion import nu_to_epsilon_prime, to_R_g_units
from ..utils.geometry import x_re_shell, mu_star_shell, x_re_ring
from ..targets import (
    CMB,
    PointSourceBehindJet,
    SSDisk,
    SphericalShellBLR,
    RingDustTorus,
)
from .kernels import compton_kernel

__all__ = ["ExternalCompton"]


class ExternalCompton:
    """class for External Compton radiation computation

    Parameters
    ----------
    blob : :class:`~agnpy.emission_regions.Blob`
        emission region and electron distribution hitting the photon target
    target : :class:`~agnpy.targets`
        class describing the target photon field    
    r : :class:`~astropy.units.Quantity`
        distance of the blob from the Black Hole (i.e. from the target photons)
    integrator : func
        function to be used for integration (default = `np.trapz`)
    """

    def __init__(self, blob, target, r=None, integrator=np.trapz):
        self.blob = blob
        # we integrate on a larger grid to account for the transformation
        # of the electron density in the reference frame of the BH
        self.gamma = self.blob.gamma_to_integrate
        self.target = target
        self.r = r
        self.integrator = integrator
        self.set_mu()
        self.set_phi()

    def set_mu(self, mu_size=100):
        self.mu_size = mu_size
        if isinstance(self.target, SSDisk):
            # in case of hte disk the mu interval does not go from -1 to 1
            r_tilde = (self.r / self.target.R_g).to_value("")
            self.mu = self.target.evaluate_mu_from_r_tilde(
                self.target.R_in_tilde, self.target.R_out_tilde, r_tilde
            )
        else:
            self.mu = np.linspace(-1, 1, self.mu_size)

    def set_phi(self, phi_size=50):
        self.phi_size = phi_size
        self.phi = np.linspace(0, 2 * np.pi, self.phi_size)

    @staticmethod
    def evaluate_sed_flux_iso_mono(
        nu,
        z,
        d_L,
        delta_D,
        mu_s,
        R_b,
        epsilon_0,
        u_0,
        n_e,
        *args,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
        mu=mu_to_integrate,
        phi=phi_to_integrate
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) for external Compton 
        on a monochromatic isotropic target photon field, for a general set of model parameters

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D : float
            Doppler factor of the relativistic outflow
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        R_b : :class:`~astropy.units.Quantity`
            size of the emitting region (spherical blob assumed)
        epsilon_0 : float
            dimensionless energy (in electron rest mass energy units) of the
            target photon field
        u_0 : :class:`~astropy.units.Quantity`
            energy density [erg cm-3] of the target photon field
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution
        mu, phi : :class:`~numpy.ndarray`
            arrays of cosine of zenith and azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversion
        epsilon_s = nu_to_epsilon_prime(nu, z)
        # multi-dimensional integration
        _gamma, _mu, _phi, _epsilon_s = axes_reshaper(gamma, mu, phi, epsilon_s)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma / delta_D, *args)
        kernel = compton_kernel(_gamma, _epsilon_s, epsilon_0, mu_s, _mu, _phi)
        integrand = N_e / np.power(_gamma, 2) * kernel
        integral_gamma = integrator(integrand, gamma, axis=0)
        integral_mu = np.trapz(integral_gamma, mu, axis=0)
        integral_phi = np.trapz(integral_mu, phi, axis=0)
        prefactor_num = (
            3 * c * sigma_T * u_0 * np.power(epsilon_s, 2) * np.power(delta_D, 3)
        )
        prefactor_denom = (
            np.power(2, 7)
            * np.power(np.pi, 2)
            * np.power(d_L, 2)
            * np.power(epsilon_0, 2)
        )
        return (prefactor_num / prefactor_denom * integral_phi).to("erg cm-2 s-1")

    def sed_flux_cmb(self, nu):
        """evaluates the flux SED for External Compton on the CMB"""
        return self.evaluate_sed_flux_iso_mono(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.mu_s,
            self.blob.R_b,
            self.target.epsilon_0,
            self.target.u_0,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            integrator=self.integrator,
            gamma=self.gamma,
            mu=self.mu,
            phi=self.phi
        )

    @staticmethod
    def evaluate_sed_flux_ps_behind_jet(
        nu,
        z,
        d_L,
        delta_D,
        mu_s,
        R_b,
        epsilon_0,
        L_0,
        r,
        n_e,
        *args,
        integrator=np.trapz,
        gamma=gamma_to_integrate
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) for external Compton 
        on a point source of photons behind the jet, for a general set of model parameters

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D : float
            Doppler factor of the relativistic outflow
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        R_b : :class:`~astropy.units.Quantity`
            size of the emitting region (spherical blob assumed)
        epsilon_0 : float
            dimensionless energy (in electron rest mass energy units) of the
            target photon field
        L_0 : :class:`~astropy.units.Quantity`
            luminosity [erg cm-3] of the point source behind the jet
        r : :class:`~astropy.units.Quantity`
            distance between the point source and the blob
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversion
        epsilon_s = nu_to_epsilon_prime(nu, z)
        # multi-dimensional integration
        _gamma, _epsilon_s = axes_reshaper(gamma, epsilon_s)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma / delta_D, *args)
        kernel = compton_kernel(_gamma, _epsilon_s, epsilon_0, mu_s, 1, 0)
        integrand = N_e / np.power(_gamma, 2) * kernel
        integral = integrator(integrand, gamma, axis=0)
        prefactor_num = (
            3 * sigma_T * L_0 * np.power(epsilon_s, 2) * np.power(delta_D, 3)
        )
        prefactor_denom = (
            np.power(2, 7)
            * np.power(np.pi, 2)
            * np.power(d_L, 2)
            * np.power(r, 2)
            * np.power(epsilon_0, 2)
        )
        return (prefactor_num / prefactor_denom * integral).to("erg cm-2 s-1")

    def sed_flux_ps_behind_jet(self, nu):
        """evaluates the flux SED for External Compton on a point source behind 
        the jet"""
        return self.evaluate_sed_flux_ps_behind_jet(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.mu_s,
            self.blob.R_b,
            self.target.epsilon_0,
            self.target.L_0,
            self.r,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            integrator=self.integrator,
            gamma=self.gamma
        )

    @staticmethod
    def evaluate_sed_flux_ss_disk(
        nu,
        z,
        d_L,
        delta_D,
        mu_s,
        R_b,
        M_BH,
        L_disk,
        eta,
        R_in,
        R_out,
        r,
        n_e,
        *args,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
        mu_size=100,
        phi=phi_to_integrate
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) for external Compton 
        on the photon field of a Shakura Sunyaev disk, for a general set of model parameters

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D: float
            Doppler factor of the relativistic outflow
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        R_b : :class:`~astropy.units.Quantity`
            size of the emitting region (spherical blob assumed)
        M_BH : :class:`~astropy.units.Quantity`
            Black Hole mass    
        L_disk : :class:`~astropy.units.Quantity`
            luminosity of the disk 
        eta : float
            accretion efficiency
        R_in : :class:`~astropy.units.Quantity` 
            inner disk radius
        R_out : :class:`~astropy.units.Quantity` 
            inner disk radius
        r : :class:`~astropy.units.Quantity`
            distance between the disk and the blob
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution
        mu_size : int
            size of the array of zenith angles to integrate over 
        phi : :class:`~numpy.ndarray`
            arrays of azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        epsilon_s = nu_to_epsilon_prime(nu, z)
        r_tilde = to_R_g_units(r, M_BH)
        R_in_tilde = to_R_g_units(R_in, M_BH)
        R_out_tilde = to_R_g_units(R_out, M_BH)
        m_dot = (L_disk / (eta * np.power(c, 2))).to("g / s")
        # multidimensional integration
        # for the disk we do not integrate mu from -1 to 1 but choose the range
        # of zenith angles subtended from a given distance
        mu = SSDisk.evaluate_mu_from_r_tilde(R_in_tilde, R_out_tilde, r_tilde, mu_size)
        _gamma, _mu, _phi, _epsilon_s = axes_reshaper(gamma, mu, phi, epsilon_s)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma / delta_D, *args)
        epsilon = SSDisk.evaluate_epsilon_mu(L_disk, M_BH, eta, _mu, r_tilde)
        phi_disk = SSDisk.evaluate_phi_disk_mu(_mu, R_in_tilde, r_tilde)
        kernel = compton_kernel(_gamma, _epsilon_s, epsilon, mu_s, _mu, _phi)
        integrand = (
            phi_disk
            / np.power(epsilon, 2)
            / _mu
            / np.power(np.power(_mu, -2) - 1, 3 / 2)
            * N_e
            / np.power(_gamma, 2)
            * kernel
        )
        integral_gamma = integrator(integrand, gamma, axis=0)
        integral_mu = np.trapz(integral_gamma, mu, axis=0)
        integral_phi = np.trapz(integral_mu, phi, axis=0)
        prefactor_num = (
            9
            * sigma_T
            * G
            * M_BH
            * m_dot
            * np.power(epsilon_s, 2)
            * np.power(delta_D, 3)
        )
        prefactor_denom = (
            np.power(2, 9) * np.power(np.pi, 3) * np.power(d_L, 2) * np.power(r, 3)
        )
        return (prefactor_num / prefactor_denom * integral_phi).to("erg cm-2 s-1")

    def sed_flux_ss_disk(self, nu):
        """evaluates the flux SED for External Compton on a [Shakura1973]_ disk"""
        return self.evaluate_sed_flux_ss_disk(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.mu_s,
            self.blob.R_b,
            self.target.M_BH,
            self.target.L_disk,
            self.target.eta,
            self.target.R_in,
            self.target.R_out,
            self.r,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            integrator=self.integrator,
            gamma=self.gamma,
            mu_size=self.mu_size,
            phi=self.phi
        )

    @staticmethod
    def evaluate_sed_flux_blr(
        nu,
        z,
        d_L,
        delta_D,
        mu_s,
        R_b,
        L_disk,
        xi_line,
        epsilon_line,
        R_line,
        r,
        n_e,
        *args,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
        mu=mu_to_integrate,
        phi=phi_to_integrate
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) for External Compton on 
        the photon field of a spherical shell BLR, for a general set of model parameters

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D: float
            Doppler factor of the relativistic outflow
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        L_disk : :class:`~astropy.units.Quantity`
            Luminosity of the disk whose radiation is being reprocessed by the BLR
        xi_line : float
            fraction of the disk radiation reprocessed by the BLR
        epsilon_line : string
            dimensionless energy of the emitted line
        R_line : :class:`~astropy.units.Quantity`
            radius of the BLR spherical shell
        r : :class:`~astropy.units.Quantity`
            distance between the Broad Line Region and the blob
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution
        mu, phi : :class:`~numpy.ndarray`
            arrays of cosine of zenith and azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        epsilon_s = nu_to_epsilon_prime(nu, z)
        # multidimensional integration
        _gamma, _mu, _phi, _epsilon_s = axes_reshaper(gamma, mu, phi, epsilon_s)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma / delta_D, *args)
        x = x_re_shell(_mu, R_line, r)
        mu_star = mu_star_shell(_mu, R_line, r)
        kernel = compton_kernel(_gamma, _epsilon_s, epsilon_line, mu_s, mu_star, _phi)
        integrand = 1 / np.power(x, 2) * N_e / np.power(_gamma, 2) * kernel
        integral_gamma = integrator(integrand, gamma, axis=0)
        integral_mu = np.trapz(integral_gamma, mu, axis=0)
        integral_phi = np.trapz(integral_mu, phi, axis=0)
        prefactor_num = (
            3
            * sigma_T
            * xi_line
            * L_disk
            * np.power(epsilon_s, 2)
            * np.power(delta_D, 3)
        )
        prefactor_denom = (
            np.power(2, 9)
            * np.power(np.pi, 3)
            * np.power(d_L, 2)
            * np.power(epsilon_line, 2)
        )
        return (prefactor_num / prefactor_denom * integral_phi).to("erg cm-2 s-1")

    def sed_flux_blr(self, nu):
        """evaluates the flux SED for External Compton on a spherical BLR"""
        return self.evaluate_sed_flux_blr(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.mu_s,
            self.blob.R_b,
            self.target.L_disk,
            self.target.xi_line,
            self.target.epsilon_line,
            self.target.R_line,
            self.r,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            integrator=self.integrator,
            gamma=self.gamma,
            mu=self.mu,
            phi=self.phi
        )

    @staticmethod
    def evaluate_sed_flux_dt(
        nu,
        z,
        d_L,
        delta_D,
        mu_s,
        R_b,
        L_disk,
        xi_dt,
        epsilon_dt,
        R_dt,
        r,
        n_e,
        *args,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
        phi=phi_to_integrate
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) for External Compton on 
        the photon field of a ring dust torus, for a general set of model parameters

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D: float
            Doppler factor of the relativistic outflow
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        L_disk : :class:`~astropy.units.Quantity`
            Luminosity of the disk whose radiation is being reprocessed by the BLR
        xi_dt : float
            fraction of the disk radiation reprocessed by the disk
        epsilon_dt : string
            peak (dimensionless) energy of the black body radiated by the torus 
        R_dt : :class:`~astropy.units.Quantity`
            radius of the ting-like torus
        r : :class:`~astropy.units.Quantity`
            distance between the Broad Line Region and the blob
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution
        mu, phi : :class:`~numpy.ndarray`
            arrays of cosine of zenith and azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        epsilon_s = nu_to_epsilon_prime(nu, z)
        # multidimensional integration
        _gamma, _phi, _epsilon_s = axes_reshaper(gamma, phi, epsilon_s)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma / delta_D, *args)
        x_re = x_re_ring(R_dt, r)
        mu = (r / x_re).to_value("")
        kernel = compton_kernel(_gamma, _epsilon_s, epsilon_dt, mu_s, mu, _phi)
        integrand = N_e / np.power(_gamma, 2) * kernel
        integral_gamma = integrator(integrand, gamma, axis=0)
        integral_phi = np.trapz(integral_gamma, phi, axis=0)
        prefactor_num = (
            3 * sigma_T * xi_dt * L_disk * np.power(epsilon_s, 2) * np.power(delta_D, 3)
        )
        prefactor_denom = (
            np.power(2, 8)
            * np.power(np.pi, 3)
            * np.power(d_L, 2)
            * np.power(x_re, 2)
            * np.power(epsilon_dt, 2)
        )
        return (prefactor_num / prefactor_denom * integral_phi).to("erg cm-2 s-1")

    def sed_flux_dt(self, nu):
        """evaluates the flux SED for External Compton on a ring dust torus"""
        return self.evaluate_sed_flux_dt(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.mu_s,
            self.blob.R_b,
            self.target.L_disk,
            self.target.xi_dt,
            self.target.epsilon_dt,
            self.target.R_dt,
            self.r,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            integrator=self.integrator,
            gamma=self.gamma,
            phi=self.phi
        )

    def sed_flux(self, nu):
        """SEDs for external Compton"""
        if isinstance(self.target, CMB):
            return self.sed_flux_cmb(nu)
        if isinstance(self.target, PointSourceBehindJet):
            return self.sed_flux_ps_behind_jet(nu)
        if isinstance(self.target, SSDisk):
            return self.sed_flux_ss_disk(nu)
        if isinstance(self.target, SphericalShellBLR):
            return self.sed_flux_blr(nu)
        if isinstance(self.target, RingDustTorus):
            return self.sed_flux_dt(nu)

    def sed_luminosity(self, nu):
        r"""Evaluates the external Compton luminosity SED
        :math:`\nu L_{\nu} \, [\mathrm{erg}\,\mathrm{s}^{-1}]`"""
        sphere = 4 * np.pi * np.power(self.blob.d_L, 2)
        return (sphere * self.sed_flux(nu)).to("erg s-1")



# plotting utilities for agnpy
from pathlib import Path
import matplotlib.pyplot as plt

# axes labels
sed_x_label = r"$\nu\,/\,\mathrm{Hz}$"
sed_y_label = r"$\nu F_{\nu}\,/\,(\mathrm{erg}\,\mathrm{cm}^{-2}\,\mathrm{s}^{-1})$"

mpl_linewidth = 1.6
mpl_rc = {
    "figure.autolayout": True,
    "figure.dpi": 150,
    "font.size": 12,
    "lines.linewidth": mpl_linewidth,
    "axes.grid": True,
    "grid.linestyle": ":",
    "grid.linewidth": 1.4,
    "axes.linewidth": mpl_linewidth,
    "xtick.major.size": 7,
    "xtick.major.width": mpl_linewidth,
    "xtick.minor.size": 4,
    "xtick.minor.width": mpl_linewidth,
    "xtick.minor.visible": True,
    "ytick.major.size": 7,
    "ytick.major.width": mpl_linewidth,
    "ytick.minor.size": 4,
    "ytick.minor.width": mpl_linewidth,
    "ytick.minor.visible": True,
}


def load_mpl_rc():
    """use the custom matplotlib rc"""
    plt.style.use(mpl_rc)


def plot_sed(nu, sed, ax=None, **kwargs):
    """plot an SED

    Parameters
    ----------
    nu: :class:`~astropy.units.Quantity`
        frequencies over which the comparison plot has to be plotted
    y_range : list of float
        lower and upper limit of the y axis limt
    comparison_range : list of float
        plot the range over which the residuals were checked 
    """
    ax = plt.gca() if ax is None else ax

    ax.loglog(nu, sed, **kwargs)
    ax.set_xlabel(sed_x_label)
    ax.set_ylabel(sed_y_label)

    if "label" in kwargs:
        ax.legend()

    return ax



# tests on spectra and emission region modules
import numpy as np
import astropy.units as u
from astropy.constants import e, c, m_e
from agnpy.emission_regions import Blob
import pytest


mec2 = m_e.to("erg", equivalencies=u.mass_energy())
# variables with _test are global and meant to be used in all tests
spectrum_norm_test = 1e-13 * u.Unit("cm-3")
gamma_min_test = 1
gamma_max_test = 1e6
# test ditctionaries for different spectra
pwl_dict_test = {
    "type": "PowerLaw",
    "parameters": {"p": 2, "gamma_min": gamma_min_test, "gamma_max": gamma_max_test},
}

bpwl_dict_test = {
    "type": "BrokenPowerLaw",
    "parameters": {
        "p1": 2.4,
        "p2": 3.4,
        "gamma_b": 1e2,
        "gamma_min": gamma_min_test,
        "gamma_max": gamma_max_test,
    },
}
lp_dict_test = {
    "type": "LogParabola",
    "parameters": {
        "p": 2.5,
        "q": 0.1,
        "gamma_0": 1e3,
        "gamma_min": gamma_min_test,
        "gamma_max": gamma_max_test,
    },
}
# blob parameters
R_b_test = 1e16 * u.cm
z_test = 0.1
delta_D_test = 10
Gamma_test = 10
B_test = 0.1 * u.G
pwl_blob_test = Blob(
    R_b_test,
    z_test,
    delta_D_test,
    Gamma_test,
    B_test,
    spectrum_norm_test,
    pwl_dict_test,
)
bpwl_blob_test = Blob(
    R_b_test,
    z_test,
    delta_D_test,
    Gamma_test,
    B_test,
    spectrum_norm_test,
    bpwl_dict_test,
)
lp_blob_test = Blob(
    R_b_test,
    z_test,
    delta_D_test,
    Gamma_test,
    B_test,
    spectrum_norm_test,
    lp_dict_test,
)
# useful for checks
Beta_test = 1 - 1 / np.power(Gamma_test, 2)
V_b_test = 4 / 3 * np.pi * np.power(R_b_test, 3)


class TestBlob:
    """class grouping all tests related to the Blob emission region"""

    def test_default_norm_type(self):
        """the default norm type should be 'integral'"""
        assert pwl_blob_test.spectrum_norm_type == "integral"

    # tests for the power-law spectrum
    # - printing test
    def test_print_pwl(self):
        print(pwl_blob_test)
        assert True

    # - tests for normalisations in cm3
    def test_pwl_integral_norm_cm3(self):
        """test if the integral norm in cm-3 is correctly set"""
        pwl_blob_test.set_spectrum(spectrum_norm_test, pwl_dict_test, "integral")
        assert u.allclose(
            pwl_blob_test.n_e_tot,
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    def test_pwl_differential_norm_cm3(self):
        """test if the differential norm in cm-3 is correctly set"""
        pwl_blob_test.set_spectrum(spectrum_norm_test, pwl_dict_test, "differential")
        assert u.allclose(
            pwl_blob_test.n_e.k_e,
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    def test_pwl_gamma_1_norm_cm3(self):
        """test if the norm at gamma = 1 in cm-3 is correctly set"""
        pwl_blob_test.set_spectrum(spectrum_norm_test, pwl_dict_test, "gamma=1")
        assert u.allclose(
            pwl_blob_test.n_e(1), spectrum_norm_test, atol=0 * u.Unit("cm-3"), rtol=1e-2
        )

    # - tests for integral normalisations in erg cm-3 and erg
    def test_pwl_integral_norm_erg_cm3(self):
        """test if the integral norm in erg cm-3 is correctly set"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        pwl_blob_test.set_spectrum(u_e, pwl_dict_test, "integral")
        assert u.allclose(
            pwl_blob_test.u_e, u_e, atol=0 * u.Unit("erg cm-3"), rtol=1e-2,
        )

    def test_pwl_integral_norm_erg(self):
        """test if the integral norm in erg is correctly set"""
        W_e = 1e48 * u.erg
        pwl_blob_test.set_spectrum(W_e, pwl_dict_test, "integral")
        assert u.allclose(pwl_blob_test.W_e, W_e, atol=0 * u.erg, rtol=1e-2)

    # tests for the broken power-law spectrum
    # - printing test
    def test_print_bpwl(self):
        print(bpwl_blob_test)
        assert True

    # - tests for normalisations in cm3
    def test_bpl_integral_norm_cm3(self):
        """test if the integral norm in cm-3 is correctly set"""
        bpwl_blob_test.set_spectrum(spectrum_norm_test, bpwl_dict_test, "integral")
        assert u.allclose(
            bpwl_blob_test.n_e_tot,
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    def test_bpl_differential_norm_cm3(self):
        """test if the differential norm in cm-3 is correctly set"""
        bpwl_blob_test.set_spectrum(spectrum_norm_test, bpwl_dict_test, "differential")
        assert u.allclose(
            bpwl_blob_test.n_e.k_e,
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    def test_bpl_gamma_1_norm_cm3(self):
        """test if the norm at gamma = 1 in cm-3 is correctly set"""
        bpwl_blob_test.set_spectrum(spectrum_norm_test, bpwl_dict_test, "gamma=1")
        assert u.allclose(
            bpwl_blob_test.n_e(1),
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    # - tests for integral normalisations in erg cm-3 and erg
    def test_bpl_integral_norm_erg_cm3(self):
        """test if the integral norm in erg cm-3 is correctly set"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        bpwl_blob_test.set_spectrum(u_e, bpwl_dict_test, "integral")
        assert u.allclose(
            bpwl_blob_test.u_e, u_e, atol=0 * u.Unit("erg cm-3"), rtol=1e-2,
        )

    def test_bpl_integral_norm_erg(self):
        """test if the integral norm in erg is correctly set"""
        W_e = 1e48 * u.erg
        bpwl_blob_test.set_spectrum(W_e, bpwl_dict_test, "integral")
        assert u.allclose(bpwl_blob_test.W_e, W_e, atol=0 * u.erg, rtol=1e-2)

    # tests for the log-parabola spectrum
    # - printing test
    def test_print_lp(self):
        print(lp_blob_test)
        assert True

    # - tests for normalisations in cm3
    def test_lp_integral_norm_cm3(self):
        """test if the integral norm in cm-3 is correctly set"""
        lp_blob_test.set_spectrum(spectrum_norm_test, lp_dict_test, "integral")
        assert u.allclose(
            lp_blob_test.n_e_tot,
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    def test_lp_differential_norm_cm3(self):
        """test if the differential norm in cm-3 is correctly set"""
        lp_blob_test.set_spectrum(spectrum_norm_test, lp_dict_test, "differential")
        assert u.allclose(
            lp_blob_test.n_e.k_e,
            spectrum_norm_test,
            atol=0 * u.Unit("cm-3"),
            rtol=1e-2,
        )

    def test_lp_gamma_1_norm_cm3(self):
        """test if the norm at gamma = 1 in cm-3 is correctly set"""
        lp_blob_test.set_spectrum(spectrum_norm_test, lp_dict_test, "gamma=1")
        assert u.allclose(
            lp_blob_test.n_e(1), spectrum_norm_test, atol=0 * u.Unit("cm-3"), rtol=1e-2
        )

    # - tests for integral normalisations in erg cm-3 and erg
    def test_lp_integral_norm_erg_cm3(self):
        """test if the integral norm in erg cm-3 is correctly set"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        lp_blob_test.set_spectrum(u_e, lp_dict_test, "integral")
        assert u.allclose(
            lp_blob_test.u_e, u_e, atol=0 * u.Unit("erg cm-3"), rtol=1e-2,
        )

    def test_lp_integral_norm_erg(self):
        """test if the integral norm in erg is correctly set"""
        W_e = 1e48 * u.erg
        lp_blob_test.set_spectrum(W_e, lp_dict_test, "integral")
        assert u.allclose(pwl_blob_test.W_e, W_e, atol=0 * u.erg, rtol=1e-2)

    # test if mismatching unit and normalisation type raises a NameError
    @pytest.mark.parametrize(
        "spectrum_norm, spectrum_norm_type",
        [
            (1e48 * u.erg, "differential"),
            (1e48 * u.erg, "gamma=1"),
            (1e2 * u.Unit("erg cm-3"), "differential"),
            (1e2 * u.Unit("erg cm-3"), "gamma=1"),
        ],
    )
    def test_non_available_norm_type(self, spectrum_norm, spectrum_norm_type):
        """check that the spectrum_norm_type 'differential' and 'gamma=1' 
        raise a NameError for a spectrum_norm in erg or erg cm-3"""
        with pytest.raises(NameError):
            pwl_blob_test.set_spectrum(spectrum_norm, pwl_dict_test, spectrum_norm_type)

    # test on blob properties
    def test_set_delta_D(self):
        pwl_blob_test.set_delta_D(Gamma=10, theta_s=20 * u.deg)
        assert np.allclose(pwl_blob_test.delta_D, 1.53804, atol=0)

    def test_set_gamma_size(self):
        pwl_blob_test.set_gamma_size(1000)
        assert len(pwl_blob_test.gamma) == 1000

    def test_N_e(self):
        """check that N_e is n_e * V_b i.e. test their ratio to be V_b"""
        pwl_blob_test.set_spectrum(spectrum_norm_test, pwl_dict_test, "differential")
        n_e = pwl_blob_test.n_e(pwl_blob_test.gamma)
        N_e = pwl_blob_test.N_e(pwl_blob_test.gamma)
        assert u.allclose(N_e / n_e, V_b_test, atol=0 * u.Unit("cm3"), rtol=1e-3)

    def test_U_B(self):
        # strip the units for convenience on this one
        U_B_expected = np.power(B_test.value, 2) / (8 * np.pi) * u.Unit("erg cm-3")
        assert np.allclose(pwl_blob_test.U_B, U_B_expected, atol=0 * u.Unit("erg cm-3"))

    def test_P_jet_e(self):
        u_e_expected = (
            mec2 * spectrum_norm_test * np.log(gamma_max_test / gamma_min_test)
        )
        P_jet_e_expected = (
            2
            * np.pi
            * np.power(R_b_test, 2)
            * Beta_test
            * np.power(Gamma_test, 2)
            * c
            * u_e_expected
        )
        assert u.allclose(
            pwl_blob_test.P_jet_e,
            P_jet_e_expected,
            atol=0 * u.Unit("erg s-1"),
            rtol=1e-2,
        )

    def test_P_jet_B(self):
        U_B_expected = np.power(B_test.value, 2) / (8 * np.pi) * u.Unit("erg cm-3")
        P_jet_B_expected = (
            2
            * np.pi
            * np.power(R_b_test, 2)
            * Beta_test
            * np.power(Gamma_test, 2)
            * c
            * U_B_expected
        )
        assert u.allclose(
            pwl_blob_test.P_jet_B,
            P_jet_B_expected,
            atol=0 * u.Unit("erg s-1"),
            rtol=1e-2,
        )



# tests on the utils module
import pytest
import numpy as np
import astropy.units as u
import agnpy.utils.math as math
import agnpy.utils.geometry as geom


twopi = 2 * np.pi


def line_loglog(x, m, n):
    """a straight line in loglog-space"""
    return x ** m * np.exp(n)


def integral_line_loglog(x_min, x_max, m, n):
    """analytical integral of the line in log-log space"""
    if np.isclose(m, -1, atol=0, rtol=1e-5):
        return np.exp(n) * np.log(x_max / x_min)
    f_low = line_loglog(x_min, m + 1, n) / (m + 1)
    f_up = line_loglog(x_max, m + 1, n) / (m + 1)
    return f_up - f_low


class TestMathUtils:
    """test uitls.math"""

    @pytest.mark.parametrize("m", np.arange(-2, 2.5, 0.5))
    @pytest.mark.parametrize("n", np.arange(-2, 2.5, 0.5))
    def test_trapz_log_log(self, m, n):
        x = np.logspace(2, 5)
        # generate syntethic power-law like data by defining a straight line
        y = line_loglog(x, m, n)
        trapz_integral = np.trapz(y, x, axis=0)
        trapz_loglog_integral = math.trapz_loglog(y, x, axis=0)
        analytical_integral = integral_line_loglog(x[0], x[-1], m, n)
        assert np.isclose(trapz_loglog_integral, analytical_integral, atol=0, rtol=0.01)


class TestUtilsGeometry:
    """test utils.geometry"""

    @pytest.mark.parametrize("R_re", [1e16 * u.cm, 1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    def test_x_re_ring_vs_x_re_ring_mu_s(self, R_re, r, uu, phi_re):
        """test consistency of both methods if mu_s ~ 1"""
        x1 = geom.x_re_ring(R_re, r + uu)
        x2 = geom.x_re_ring_mu_s(R_re, r, phi_re, uu, 0.9999)
        assert np.isclose(x1, x2, atol=0, rtol=0.01)

    @pytest.mark.parametrize("R_re", [1e16 * u.cm, 1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    def test_x_re_ring_mu_s_perp(self, R_re, r, uu, phi_re):
        """test if the calculations are fine for mu_s ~=0"""
        mu_s = 0.01
        x1 = geom.x_re_ring_mu_s(R_re, r, phi_re, uu, mu_s)
        x2 = np.sqrt(
            r ** 2 + (uu - R_re * np.cos(phi_re)) ** 2 + (R_re * np.sin(phi_re)) ** 2
        )
        assert np.isclose(x1, x2, atol=0, rtol=0.01)

    @pytest.mark.parametrize("R_re", [1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0 + 0.1, 2 * np.pi - 0.1, 5))
    def test_phi_mu_re_ring_1(self, R_re, r, uu, phi_re):
        """test for mu_s = 1"""
        mu_s = 0.99999
        x_re = geom.x_re_ring_mu_s(R_re, r, phi_re, uu, mu_s)
        phi, mu = geom.phi_mu_re_ring(R_re, r, phi_re, uu, mu_s)
        mu2 = (r + uu) / x_re
        assert np.isclose(mu, mu2, atol=0.01, rtol=0)

        # there is a pi of difference between phi and phi_re
        dphi = np.mod(phi.value - phi_re + twopi, twopi)
        # this one fails unless mu_s is very close to one, or uu << R_re
        assert np.isclose(dphi, np.pi, atol=0.1, rtol=0)

    @pytest.mark.parametrize("R_re", [1e16 * u.cm, 1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    def test_phi_mu_re_ring_0(self, R_re, r, uu, phi_re):
        """test for mu_s =~ 0"""
        mu_s = 0.001
        x_re = geom.x_re_ring_mu_s(R_re, r, phi_re, uu, mu_s)
        phi, mu = geom.phi_mu_re_ring(R_re, r, phi_re, uu, mu_s)
        mu2 = r / x_re
        assert np.isclose(mu, mu2, atol=0.01, rtol=0)

        phi2 = np.arctan2(-R_re * np.sin(phi_re), uu - R_re * np.cos(phi_re)).value

        # here we add on purpose pi to avoid slightly negative differences
        dphi = np.mod(phi.value - phi2 + 3 * np.pi, twopi)
        assert np.isclose(dphi, np.pi, atol=0.1, rtol=0)

    @pytest.mark.parametrize("R_re", [1e16 * u.cm, 1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [1.0e19 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    @pytest.mark.parametrize("mu_s", np.linspace(0, 1, 5))
    def test_phi_mu_infty(self, R_re, r, uu, phi_re, mu_s):
        """test that for uu >> R_re for mu_s =~ 0"""
        phi, mu = geom.phi_mu_re_ring(R_re, r, phi_re, uu, mu_s)
        cospsi = geom.cos_psi(mu_s, mu, phi)
        assert np.isclose(cospsi, 1, atol=0.01, rtol=0)

    @pytest.mark.parametrize("R_re", [1e16 * u.cm, 1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    @pytest.mark.parametrize("mu_s", np.linspace(0, 1, 5))
    def test_x_re_shell_mu_s_vs_ring(self, R_re, r, phi_re, uu, mu_s):
        """Test that for mu_re=0. x_re_shell_mu_s gives the same results as 
        x_re_ring_mu_s"""
        x_ring = geom.x_re_ring_mu_s(R_re, r, phi_re, uu, mu_s)
        mu_re = 0.01
        x_shell = geom.x_re_shell_mu_s(R_re, r, phi_re, mu_re, uu, mu_s)
        assert np.isclose(x_shell, x_ring, atol=0, rtol=0.01)

    @pytest.mark.parametrize("R_re", [1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [1.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    @pytest.mark.parametrize("mu_re", np.linspace(-1, 1, 4))
    def test_x_re_shell_mu_s_vs_on_axis(self, R_re, r, phi_re, mu_re, uu):
        """Test that for mu_s=1. x_re_shell_mu_s gives the same results as 
        x_re_shell"""
        mu_s = 0.9999
        x_on_axis = geom.x_re_shell(mu_re, R_re, r + uu)
        x_shell = geom.x_re_shell_mu_s(R_re, r, phi_re, mu_re, uu, mu_s)
        assert np.isclose(x_shell, x_on_axis, atol=0, rtol=0.01)

    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    def test_x_re_shell_mu_s(self, phi_re):
        """Test that for a simple case that  x_re_shell_mu_s gives correct results"""
        R_re = 1e17 * u.cm
        r = 3e17 * u.cm
        uu = 5e17 * u.cm
        mu_re = 0.999
        mu_s = 0.001
        x_true = np.sqrt((r - R_re) ** 2 + uu ** 2)
        x_shell = geom.x_re_shell_mu_s(R_re, r, phi_re, mu_re, uu, mu_s)
        assert np.isclose(x_shell, x_true, atol=0, rtol=0.01)

    @pytest.mark.parametrize("R_re", [1e16 * u.cm, 1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    @pytest.mark.parametrize("mu_s", np.linspace(0, 1, 5))
    def test_phi_mu_re_shell_vs_ring(self, R_re, r, phi_re, uu, mu_s):
        """Test that for mu_re=0 phi_mu_re_shell gives the same results as 
        phi_mu_re_ring"""
        phi_ring, mu_ring = geom.phi_mu_re_ring(R_re, r, phi_re, uu, mu_s)
        mu_re = 0.001
        phi_shell, mu_shell = geom.phi_mu_re_shell(R_re, r, phi_re, mu_re, uu, mu_s)
        assert np.isclose(mu_shell, mu_ring, atol=0.01, rtol=0)

        # here we add on purpose pi to avoid slightly negative differences
        dphi = np.mod(phi_shell.value - phi_ring.value + 3 * np.pi, twopi)
        assert np.isclose(dphi, np.pi, atol=0.01, rtol=0)

    @pytest.mark.parametrize("R_re", [1e17 * u.cm])
    @pytest.mark.parametrize("r", [3.0e16 * u.cm, 2.0e17 * u.cm])
    @pytest.mark.parametrize("uu", [5.0e16 * u.cm, 4.0e17 * u.cm])
    @pytest.mark.parametrize("phi_re", np.linspace(0, 2 * np.pi, 5))
    @pytest.mark.parametrize("mu_s", np.linspace(0, 0.9, 5))
    def test_phi_mu_re_shell(self, R_re, r, phi_re, uu, mu_s):
        """Test that for mu_re=1 phi_mu_re_shell gives correct result"""
        mu_re = 0.99999
        phi_shell, mu_shell = geom.phi_mu_re_shell(R_re, r, phi_re, mu_re, uu, mu_s)
        phi_true = 0  # unless mu_s = 1 in which case phi_true can be whatever
        dx = uu * np.sqrt(1 - mu_s ** 2)
        dz = r - R_re + uu * mu_s
        mu_true = dz / np.sqrt(dx ** 2 + dz ** 2)
        assert np.isclose(mu_shell, mu_true, atol=0.01, rtol=0)
        print(phi_true, phi_shell, R_re, r, phi_re, uu, mu_s)

        # here we add on purpose pi to avoid slightly negative differences
        dphi = np.mod(phi_shell.value - phi_true + 3 * np.pi, twopi)
        assert np.isclose(dphi, np.pi, atol=0.03, rtol=0)



# test on absorption module
import pytest
import numpy as np
import astropy.units as u
from astropy.constants import m_e, c, M_sun, sigma_T, k_B
from astropy.coordinates import Distance
from pathlib import Path
from agnpy.emission_regions import Blob
from agnpy.synchrotron import Synchrotron
from agnpy.targets import PointSourceBehindJet, SSDisk, SphericalShellBLR, RingDustTorus
from agnpy.absorption import Absorption, EBL, ebl_files_dict
from agnpy.utils.math import axes_reshaper
from .utils import (
    make_comparison_plot,
    extract_columns_sample_file,
    check_deviation,
)
import matplotlib.pyplot as plt

mec2 = m_e.to("erg", equivalencies=u.mass_energy())

agnpy_dir = Path(__file__).parent.parent
# where to read sampled files
data_dir = agnpy_dir / "data"
# where to save figures
figures_dir = agnpy_dir.parent / "crosschecks/figures/absorption"
for subdir in ["ebl", "disk", "blr", "dt"]:
    Path(figures_dir / subdir).mkdir(parents=True, exist_ok=True)


class TestAbsorption:
    """class grouping all tests related to the Absorption class"""

    @pytest.mark.parametrize("r", ["1e-1", "1e0", "1e1", "1e2"])
    def test_absorption_disk_reference_tau(self, r):
        """test agnpy gamma-gamma optical depth for Disk against the one in 
        Figure 14 of Finke 2016"""
        # reference tau
        E_ref, tau_ref = extract_columns_sample_file(
            f"{data_dir}/reference_taus/finke_2016/figure_14_left/tau_SSdisk_r_{r}_R_Ly_alpha.txt",
            "GeV",
        )
        nu_ref = E_ref.to("Hz", equivalencies=u.spectral())
        # target
        M_BH = 1.2 * 1e9 * M_sun.cgs
        L_disk = 2e46 * u.Unit("erg s-1")
        eta = 1 / 12
        R_in = 6
        R_out = 200
        disk = SSDisk(M_BH, L_disk, eta, R_in, R_out, R_g_units=True)
        R_Ly_alpha = 1.1e17 * u.cm
        _r = float(r) * R_Ly_alpha
        # recompute the tau, use the full energy range of figure 14
        z = 0.859
        abs_disk = Absorption(disk, _r, z)
        tau_agnpy = abs_disk.tau(nu_ref)
        # comparison plot
        make_comparison_plot(
            nu_ref,
            tau_agnpy,
            tau_ref,
            "agnpy",
            "Figure 14, Finke (2016)",
            f"Absorption Shakura Sunyaev Disk, r = {r} R(Ly alpha)",
            f"{figures_dir}/disk/tau_disk_comparison_r_{r}_R_Ly_alpha_figure_14_finke_2016.png",
            "tau",
            y_range=[1e-10, 1e5],
        )
        assert True

    @pytest.mark.parametrize(
        "r,nu_min", [("1e-1", 5e24), ("1e0.5", 2e26), ("1e1", 2.3e27)]
    )
    def test_absorption_blr_reference_tau(self, r, nu_min):
        """test agnpy gamma-gamma optical depth for a Lyman alpha BLR against 
        the one in Figure 14 of Finke 2016"""
        # reference tau
        E_ref, tau_ref = extract_columns_sample_file(
            f"{data_dir}/reference_taus/finke_2016/figure_14_left/tau_BLR_Ly_alpha_r_{r}_R_Ly_alpha.txt",
            "GeV",
        )
        # Finke's frequencies are not corrected for the redshift
        z = 0.859
        nu_ref = E_ref.to("Hz", equivalencies=u.spectral()) / (1 + z)
        # target
        L_disk = 2 * 1e46 * u.Unit("erg s-1")
        xi_line = 0.024
        R_Ly_alpha = 1.1 * 1e17 * u.cm
        blr = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_Ly_alpha)
        # parse the string providing the distance in units of R_Ly_alpha
        # numbers as 1e1.5 cannot be directly converted to float
        num = r.split("e")
        Ly_alpha_units = (float(num[0]) * 10) ** (float(num[-1]))
        _r = Ly_alpha_units * R_Ly_alpha
        # recompute the tau, use the full energy range of figure 14
        abs_blr = Absorption(blr, _r, z)
        tau_agnpy = abs_blr.tau(nu_ref)
        # check in a restricted energy range
        nu_range = [nu_min, 3e28] * u.Hz
        make_comparison_plot(
            nu_ref,
            tau_agnpy,
            tau_ref,
            "agnpy",
            "Figure 14, Finke (2016)",
            f"Absorption on Spherical Shell BLR, r = {Ly_alpha_units:.2e} R(Ly alpha)",
            f"{figures_dir}/blr/tau_blr_Ly_alpha_comprison_r_{r}_R_Ly_alpha_figure_14_finke_2016.png",
            "tau",
            comparison_range=nu_range.to_value("Hz"),
            y_range=[1e-6, 1e3],
        )
        # requires that the SED points deviate less than 35 % from those of the reference figure
        assert check_deviation(nu_ref, tau_agnpy, tau_ref, 0.35, nu_range)

    @pytest.mark.parametrize(
        "r,nu_min", [("1e-1", 3.3e26), ("1e0", 3.3e26), ("1e1", 3.3e26), ("1e2", 8e26)]
    )
    def test_absorption_dt_reference_tau(self, r, nu_min):
        """test agnpy gamma-gamma optical depth for DT against the one in 
        Figure 14 of Finke 2016"""
        # reference tau
        E_ref, tau_ref = extract_columns_sample_file(
            f"{data_dir}/reference_taus/finke_2016/figure_14_left/tau_DT_r_{r}_R_Ly_alpha.txt",
            "GeV",
        )
        nu_ref = E_ref.to("Hz", equivalencies=u.spectral())
        # target
        L_disk = 2 * 1e46 * u.Unit("erg s-1")
        T_dt = 1e3 * u.K
        csi_dt = 0.1
        dt = RingDustTorus(L_disk, csi_dt, T_dt)
        R_Ly_alpha = 1.1 * 1e17 * u.cm
        # parse the string providing the distance in units of R_Ly_alpha
        # numbers as 1e1.5 cannot be directly converted to float
        num = r.split("e")
        Ly_alpha_units = (float(num[0]) * 10) ** (float(num[-1]))
        _r = Ly_alpha_units * R_Ly_alpha
        # recompute the tau, use the full energy range of figure 14
        z = 0.859
        abs_dt = Absorption(dt, _r, z)
        tau_agnpy = abs_dt.tau(nu_ref)
        # check in a restricted energy range
        nu_range = [nu_min, 3e28] * u.Hz
        make_comparison_plot(
            nu_ref,
            tau_agnpy,
            2 * tau_ref,
            "agnpy",
            "Figure 14, Finke (2016)",
            f"Absorption on Dust Torus, r = {Ly_alpha_units:.2e} R(Ly alpha)",
            f"{figures_dir}/dt/tau_dt_comprison_r_{r}_R_Ly_alpha_figure_14_finke_2016.png",
            "tau",
            comparison_range=nu_range.to_value("Hz"),
            y_range=[1e-6, 1e3],
        )
        # requires that the SED points deviate less than 20 % from those of the reference figure
        assert check_deviation(nu_ref, tau_agnpy, 2 * tau_ref, 0.20, nu_range)

    def test_abs_blr_vs_point_source(self):
        """check if in the limit of large distances the gamma-gamma optical depth 
        on the BLR tends to the one of a point-like source approximating it"""
        # broad line region
        L_disk = 2e46 * u.Unit("erg s-1")
        xi_line = 0.024
        R_line = 1e17 * u.cm
        blr = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_line)
        r = 1e20 * u.cm
        # point like source approximating the blr
        ps_blr = PointSourceBehindJet(blr.xi_line * L_disk, blr.epsilon_line)
        # absorption, consider a small viewing angle for this case
        z = 0.859
        theta_s = np.deg2rad(10)
        abs_blr = Absorption(blr, r, z, mu_s=np.cos(theta_s))
        abs_ps_blr = Absorption(ps_blr, r, z, mu_s=np.cos(theta_s))
        # taus
        E = np.logspace(2, 6) * u.GeV
        nu = E.to("Hz", equivalencies=u.spectral())
        tau_blr = abs_blr.tau(nu)
        tau_ps_blr = abs_ps_blr.tau(nu)
        # sed comparison plot
        make_comparison_plot(
            nu,
            tau_ps_blr,
            tau_blr,
            "point source approximating the BLR",
            "spherical shell BLR",
            "Absorption on Spherical Shell BLR, "
            + r"$r = 10^{20}\,{\rm cm} \gg R({\rm Ly\alpha}),\,\theta_s=10^{\circ}$",
            f"{figures_dir}/blr/tau_blr_point_source_comparison.png",
            "tau",
        )
        # requires a 10% deviation from the two SED points
        assert check_deviation(nu, tau_blr, tau_ps_blr, 0.1)

    def test_abs_dt_vs_point_source(self):
        """check if in the limit of large distances the gamma-gamma optical depth 
        on the DT tends to the one of a point-like source approximating it"""
        # dust torus
        L_disk = 2e46 * u.Unit("erg s-1")
        T_dt = 1e3 * u.K
        csi_dt = 0.1
        dt = RingDustTorus(L_disk, csi_dt, T_dt)
        r = 1e22 * u.cm
        # point like source approximating the dt
        ps_dt = PointSourceBehindJet(dt.xi_dt * L_disk, dt.epsilon_dt)
        # absorption
        z = 0.859
        theta_s = np.deg2rad(10)
        abs_dt = Absorption(dt, r, z, mu_s=np.cos(theta_s))
        abs_ps_dt = Absorption(ps_dt, r, z, mu_s=np.cos(theta_s))
        # taus
        E = np.logspace(2, 6) * u.GeV
        nu = E.to("Hz", equivalencies=u.spectral())
        tau_dt = abs_dt.tau(nu)
        tau_ps_dt = abs_ps_dt.tau(nu)
        make_comparison_plot(
            nu,
            tau_ps_dt,
            tau_dt,
            "point source approximating the DT",
            "ring dust torus",
            "Absorption on Ring Dust Torus, "
            + r"$r = 10^{22}\,{\rm cm} \gg R_{\rm dt},\,\theta_s=10^{\circ}$",
            f"{figures_dir}/dt/tau_dt_point_source_comparison.png",
            "tau",
        )
        # requires a 10% deviation from the two SED points
        assert check_deviation(nu, tau_dt, tau_ps_dt, 0.1)


def sigma_pp(b):
    """pair production cross section"""
    return (
        sigma_T
        * 3.0
        / 16.0
        * (1 - b ** 2)
        * (2 * b * (b ** 2 - 2) + (3 - b ** 4) * np.log((1 + b) / (1 - b)))
    )


class TestAbsorptionMuS:
    """test absorption functions for calculations with mu_s !=1"""

    def test_tau_dt_mu_s_simple(self):
        """
        order of magnitude test comparing with simplified calculations
        the case of a perpendicularly moving photon starting at r~0.5 * R_re
        """
        r = 1.0e17 * u.cm  # distance at which the photon starts
        mu_s = 0.0  # angle of propagation

        L_disk = 2e46 * u.Unit("erg s-1")
        xi_DT = 0.1
        temp = 1000 * u.K
        R_DT = 2 * r  # radius of DT: assume the same as the distance r
        dt = RingDustTorus(L_disk, xi_DT, temp, R_dt=R_DT)

        nu_ref = np.logspace(26, 32, 120) * u.Hz

        # absorption at mu_s
        abs_dt_mu_s = Absorption(dt, r, z=0, mu_s=mu_s)
        tau_dt_mu_s = abs_dt_mu_s.tau(nu_ref)

        eps = (2.7 * temp * k_B).to("eV")  # energy of soft photons
        # soft photon density
        nph = (L_disk * xi_DT / (4 * np.pi * (r ** 2 + R_DT ** 2)) / c / eps).to("cm-3")

        E = nu_ref.to("eV", equivalencies=u.spectral())
        cospsi = 0  # assume perpendicular scattering
        beta2 = 1 - 2 * mec2 ** 2 / (E * eps * (1 - cospsi))
        beta2[beta2 < 0] = 0  # below the threshold
        # for tau calculations we assume that gamma ray moves
        # roughtly the characteristic distance of ~R_DT
        tau_my = (sigma_pp(np.sqrt(beta2)) * nph * R_DT * (1 - cospsi)).to("")

        max_agnpy = max(tau_dt_mu_s)
        max_my = max(tau_my)
        max_pos_agnpy = nu_ref[np.argmax(tau_dt_mu_s)]
        max_pos_my = nu_ref[np.argmax(tau_my)]
        # very rough calculations so allowing for
        # 15% accuracy in peak maximum and 30% in peak position
        assert np.isclose(max_agnpy, max_my, atol=0, rtol=0.15)
        assert np.isclose(max_pos_agnpy, max_pos_my, atol=0, rtol=0.3)

    @pytest.mark.parametrize("mu_s", np.linspace(0, 0.9, 5))
    def test_tau_dt_mu_s_far(self, mu_s):
        """
        comparing the DT absorption for mu_s !=1 with point source simplification
        """
        L_disk = 2e46 * u.Unit("erg s-1")
        xi_DT = 0.1
        temp = 1000 * u.K
        R_DT = 1.0e17 * u.cm  # radius of DT
        dt = RingDustTorus(L_disk, xi_DT, temp, R_dt=R_DT)

        r = 10 * R_DT  # distance at which the photon starts

        nu_ref = np.logspace(26, 32, 60) * u.Hz

        # absorption at mu_s
        abs_dt_mu_s = Absorption(dt, r, z=0, mu_s=mu_s)
        tau_dt_mu_s = abs_dt_mu_s.tau(nu_ref)

        uu = np.logspace(-5, 5, 100) * r
        _u, _nu_ref = axes_reshaper(uu, nu_ref)

        eps = (2.7 * temp * k_B).to("eV")  # energy of soft photons
        # distance from the DT
        _x = np.sqrt(r * r + _u * _u + 2 * mu_s * _u * r)

        # soft photon density
        _nph = (L_disk * xi_DT / (4 * np.pi * _x ** 2) / c / eps).to("cm-3")

        _E = _nu_ref.to("eV", equivalencies=u.spectral())
        _cospsi = (_u ** 2 + _x ** 2 - r ** 2) / (2 * _u * _x)

        _beta2 = 1 - 2 * mec2 ** 2 / (_E * eps * (1 - _cospsi))
        _beta2[_beta2 < 0] = 0

        integrand = sigma_pp(np.sqrt(_beta2)) * _nph * (1 - _cospsi)
        tau_my = (np.trapz(integrand, uu, axis=0)).to("")

        print(tau_my / tau_dt_mu_s)

        max_agnpy = max(tau_dt_mu_s)
        max_my = max(tau_my)
        max_pos_agnpy = nu_ref[np.argmax(tau_dt_mu_s)]
        max_pos_my = nu_ref[np.argmax(tau_my)]

        # if r>>R_re this should be pretty precise, allowing for 10% accuracy
        assert np.isclose(max_agnpy, max_my, atol=0, rtol=0.1)
        assert np.isclose(max_pos_agnpy, max_pos_my, atol=0, rtol=0.1)

    @pytest.mark.parametrize("r_to_R", ["0.11", "10."])
    def test_abs_blr_mu_s_vs_on_axis(self, r_to_R):
        """check if the codes computing absorption on BLR for mu_s = 1 and !=1 cases are consistent """
        # broad line region
        L_disk = 2e46 * u.Unit("erg s-1")
        xi_line = 0.024
        R_line = 1e17 * u.cm
        blr = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_line)
        r = r_to_R * R_line
        z = 0.859

        abs_blr = Absorption(blr, r, z, mu_s=0.9999)
        abs_blr_on_axis = Absorption(blr, r, z)
        abs_blr.set_l(50)
        abs_blr_on_axis.set_l(50)

        # taus
        E = np.logspace(0, 6) * u.GeV
        nu = E.to("Hz", equivalencies=u.spectral())
        tau_blr = abs_blr.tau(nu)
        tau_blr_on_axis = abs_blr_on_axis.tau(nu)
        # sed comparison plot
        make_comparison_plot(
            nu,
            tau_blr_on_axis,
            tau_blr,
            "on-axis calculations",
            "general",
            "Absorption on Spherical Shell BLR, " + r"$r/R_{\rm line}=$" + f"{r_to_R}",
            f"{figures_dir}/blr/tau_blr_on_axis_vs_general_r_{r_to_R}_R_line_comparison.png",
            "tau",
        )

        # only check in there range with measurable absorption
        xmin = min(nu[tau_blr_on_axis > 1.0e-4])
        xmax = max(nu[tau_blr_on_axis > 1.0e-4])
        xrange = (xmin, xmax)
        print(xrange)

        # close to the threshold there are some differences up to ~25%
        # which are probably due to numerical uncertainties in the integrals
        assert check_deviation(nu, tau_blr, tau_blr_on_axis, 0.25, x_range=xrange)

    @pytest.mark.parametrize("r_R_line", [1, 1.0e5 ** (-10.0 / 49)])
    def test_tau_blr_Rline(self, r_R_line):
        """
        Checks if absorption on BLR works also fine
        if one of the integration points falls on R_line
        """
        L_disk = 2 * 1e46 * u.Unit("erg s-1")
        xi_line = 0.024
        R_line = 1.1 * 1e17 * u.cm
        blr = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_line)

        abs_blr = Absorption(blr, r=r_R_line * R_line)

        # reference (without problem)
        abs_blr0 = Absorption(blr, r=r_R_line * R_line * 1.001)

        nu = np.logspace(22, 28) * u.Hz
        tau_blr = abs_blr.tau(nu)
        tau_blr0 = abs_blr0.tau(nu)

        # the current integrals are not very precise, and this includes
        # tricky part to integrate, so allowing for rather large error margin
        assert np.allclose(tau_blr, tau_blr0, atol=0.01, rtol=0.04)

    def find_r_for_x_cross(mu_s, ipoint, npoints):
        """
        Finding which r should be used to fall with one of the points on top of BLR sphere.
        """
        # u = alpha * r
        alpha = 1.0e-5 * 1.0e10 ** (ipoint / (npoints - 1.0))
        return 1 / np.sqrt(alpha ** 2 + 2 * alpha * mu_s + 1)

    @pytest.mark.parametrize("r_R_line", [1, find_r_for_x_cross(0.99, 60, 100)])
    def test_tau_blr_mus_Rline(self, r_R_line):
        """
        Checks if absorption on BLR works also fine
        if one of the integration points falls on R_line
        """
        L_disk = 2 * 1e46 * u.Unit("erg s-1")
        xi_line = 0.024
        R_line = 1.1 * 1e17 * u.cm
        blr = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_line)

        abs_blr = Absorption(blr, r=r_R_line * R_line, mu_s=0.99)

        # reference (without problem)
        abs_blr0 = Absorption(blr, r=r_R_line * R_line * 1.001, mu_s=0.99)

        nu = np.logspace(22, 28) * u.Hz
        tau_blr = abs_blr.tau(nu)
        tau_blr0 = abs_blr0.tau(nu)

        # the current integrals are not very precise, and this includes
        # tricky part to integrate, so allowing for rather large error margin
        assert np.allclose(tau_blr, tau_blr0, atol=0.01, rtol=1.04)

    def test_absorption_failing_without_r(self):
        """Tests that you cannot run absorption on "fixed" targets
        (like BLR, DT or the disk)  without specifying distance"""

        L_disk = 2 * 1e46 * u.Unit("erg s-1")
        xi_line = 0.024
        R_line = 1.1 * 1e17 * u.cm
        blr = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_line)

        # no distance specified, should fail
        with pytest.raises(ValueError):
            abs_blr = Absorption(blr)

    def test_tau_on_synchrotron_compare_with_delta_approximation(self):
        """Compare the output of the calculation of absorption of gamma rays
        in synchrotron radiation and compare it with simplified delta approximation.
        """

        # create a test blob
        r_b = 1.0e16 * u.cm
        z = 2.01
        delta_D = 10.1
        Gamma = 10.05
        B0 = 1.1 * u.G
        gmin = 10
        gbreak = 1.0e3
        gmax = 1.0e5
        spectrum_norm = 0.1 * u.Unit("erg cm-3")
        parameters = {
            "p1": 1.5,
            "p2": 2.5,
            "gamma_b": gbreak,
            "gamma_min": gmin,
            "gamma_max": gmax,
        }
        spectrum_dict = {"type": "BrokenPowerLaw", "parameters": parameters}
        blob = Blob(r_b, z, delta_D, Gamma, B0, spectrum_norm, spectrum_dict)

        nu_tau = np.logspace(22, 34, 100) * u.Hz  # for absorption calculations
        e_tau = nu_tau.to("eV", equivalencies=u.spectral())

        # full calculations
        absorb = Absorption(blob)
        tau = absorb.tau(nu_tau)

        # now simplified calculations using Eq. 37 of Finke 2008
        mec2 = (m_e * c ** 2).to("eV")
        eps1 = e_tau / mec2
        eps1p = eps1 * (1 + z) / blob.delta_D
        eps_bar = 2 * blob.delta_D ** 2 / (1 + z) ** 2 / eps1
        nu_bar = (eps_bar * mec2).to("Hz", equivalencies=u.spectral())
        synch = Synchrotron(blob, ssa=True)
        synch_sed_ebar = synch.sed_flux(nu_bar)

        tau_delta = (
            synch_sed_ebar
            * sigma_T
            * Distance(z=z) ** 2
            * eps1p
            / (2 * m_e * c ** 3 * blob.R_b * blob.delta_D ** 4)
        )
        tau_delta = tau_delta.to("")

        # the delta approximation does not work well with sharp cut-offs of synchrotron SED.
        # We first find the peak of the synchr SED
        maxidx = np.where(synch_sed_ebar == np.amax(synch_sed_ebar))[0][0]
        # and only take energies below the peak, (note that energies are ordered
        # in reversed order), and  take only those that are at least 1% of the peak
        idxs = synch_sed_ebar[maxidx:] > 1.0e-2 * synch_sed_ebar[maxidx]

        # if there are not many points something went wrong with the test
        assert sum(idxs) > 10

        # the agreement in the middle range is pretty good, but at the edges
        # of energy range it spoils very fast, so only rough agreement is checked
        # (0.6 in the log space)
        assert np.allclose(
            np.log(tau)[maxidx:][idxs], np.log(tau_delta)[maxidx:][idxs], atol=0.6
        )

    def test_absorption_pointlike_and_homogeneous(self):
        """Simple test for checking the attenuation factors in both considered cases."""

        blob = Blob()
        absorb = Absorption(blob)

        nu_tau = np.logspace(22, 34, 100) * u.Hz  # for absorption calculations

        tau = absorb.tau(nu_tau)
        abs_pointlike = absorb.absorption(nu_tau)
        abs_homogeneous = absorb.absorption_homogeneous(nu_tau)
        # select good points (to avoid division by zero)
        idxs = tau > 1.0e-3
        # if this fails something is wrong with the test itself (e.g. default parameters of the blob)
        assert sum(idxs) > 10

        # the formulas reproduce what should be in the function, so the agreement should be down to numerical accuracy
        assert np.allclose(np.exp(-tau), abs_pointlike, atol=1.0e-5)
        assert np.allclose((1 - np.exp(-tau)) / tau, abs_homogeneous, atol=1.0e-5)


class TestEBL:
    """class grouping all tests related to the EBL class"""

    @pytest.mark.parametrize("model", ["franceschini", "finke", "dominguez"])
    @pytest.mark.parametrize("z", [0.5, 1.5])
    def test_correct_interpolation(self, model, z):
        # define the ebl model, evaluate it at the reference energies
        ebl = EBL(model)
        nu_ref = ebl.energy_ref.to("Hz", equivalencies=u.spectral())
        absorption = ebl.absorption(z, nu_ref)
        # find in the reference values the spectra for this redshift
        z_idx = np.abs(z - ebl.z_ref).argmin()
        absorption_ref = ebl.values_ref[z_idx]
        make_comparison_plot(
            nu_ref,
            absorption,
            absorption_ref,
            "agnpy interpolation",
            "data",
            f"EBL absorption, {model} model, z = {z}",
            f"{figures_dir}/ebl/ebl_abs_interp_comparison_{model}_z_{z}.png",
            "abs.",
        )
        # requires a 1% deviation from the two SED points
        assert check_deviation(nu_ref, absorption_ref, absorption, 0.01)



# integration kernels for inverse Compton evaluation
import numpy as np
from ..utils.math import log
from ..utils.geometry import cos_psi

__all__ = ["F_c", "isotropic_kernel", "get_gamma_min", "compton_kernel"]


def F_c(q, gamma_e):
    """isotropic Compton kernel, Eq. 6.75 in [DermerMenon2009]_, Eq. 10 in [Finke2008]_"""
    term_1 = 2 * q * log(q)
    term_2 = (1 + 2 * q) * (1 - q)
    term_3 = 1 / 2 * np.power(gamma_e * q, 2) / (1 + gamma_e * q) * (1 - q)
    return term_1 + term_2 + term_3


def isotropic_kernel(gamma, epsilon, epsilon_s):
    """Compton kernel for isotropic nonthermal electrons scattering photons of 
    an isotropic external radiation field.
    Integrand of Eq. 6.74 in [DermerMenon2009]_.
    Parameters
    ----------
    gamma : :class:`~numpy.ndarray`
        Lorentz factors of the electrons distribution
    epsilon : :class:`~numpy.ndarray`
        dimensionless energies (in electron rest mass units) of the target photons
    epsilon_s : :class:`~numpy.ndarray`
        dimensionless energies (in electron rest mass units) of the scattered photons
    """
    gamma_e = 4 * gamma * epsilon
    q = (epsilon_s / gamma) / (gamma_e * (1 - epsilon_s / gamma))
    q_min = 1 / (4 * np.power(gamma, 2))
    return np.where((q_min <= q) * (q <= 1), F_c(q, gamma_e), 0)


def get_gamma_min(epsilon_s, epsilon, mu_s, mu, phi):
    """minimum Lorentz factor for Compton integration, 
    Eq. 29 in [Dermer2009]_, Eq. 38 in [Finke2016]_."""
    sqrt_term = np.sqrt(1 + 2 / (epsilon * epsilon_s * (1 - cos_psi(mu_s, mu, phi))))
    return epsilon_s / 2 * (1 + sqrt_term)


def compton_kernel(gamma, epsilon_s, epsilon, mu_s, mu, phi):
    """angle dependent Compton kernel:
    Eq. 26-27 in [Dermer2009]_.
    Parameters
    ----------
    gamma : :class:`~numpy.ndarray`
        Lorentz factors of the electrons distribution
    epsilon : :class:`~numpy.ndarray`
        dimesnionless energies (in electron rest mass units) of the target photons
    epsilon_s : :class:`~numpy.ndarray`
        dimensionless energies (in electron rest mass units) of the scattered photons
    mu_s : float
        cosine of the zenith angle of the blob w.r.t the jet
    mu : :class:`~numpy.ndarray` or float
        (array of) cosine of the zenith angle subtended by the target
    phi : :class:`~numpy.ndarray` or float
        (array of) of the azimuth angle subtended by the target
    """
    epsilon_bar = gamma * epsilon * (1 - cos_psi(mu_s, mu, phi))
    y = 1 - epsilon_s / gamma
    y_1 = -(2 * epsilon_s) / (gamma * epsilon_bar * y)
    y_2 = np.power(epsilon_s, 2) / np.power(gamma * epsilon_bar * y, 2)
    values = y + 1 / y + y_1 + y_2
    gamma_min = get_gamma_min(epsilon_s, epsilon, mu_s, mu, phi)
    values = np.where(gamma >= gamma_min, y + 1 / y + y_1 + y_2, 0)
    return values



# module containing the electron spectra
import numpy as np
import astropy.units as u
from ..utils.math import trapz_loglog
from ..utils.conversion import mec2

__all__ = [
    "ElectronDistribution",
    "PowerLaw",
    "ExpCutoffPowerLaw",
    "BrokenPowerLaw",
    "LogParabola",
]


class ElectronDistribution:
    """Base class grouping common functionalities to be used by all electron 
    distributions. Choose the function to be used for integration. The
    default is :class:`~numpy.trapz`"""

    def __init__(self, integrator=np.trapz):
        self.integrator = integrator

    @staticmethod
    def general_integral(
        self, gamma_low, gamma_up, gamma_power=0, integrator=np.trapz, **kwargs
    ):
        """integral of the electron distribution over the range gamma_low, 
        gamma_up for a general set of parameters

        Parameters
        ----------
        gamma_low : float
            lower integration limit
        gamma_up : float
            higher integration limit
        gamma_power : int
            power of gamma to raise the electron distribution before integration
        integrator: func
            function to be used for integration, default is :class:`~numpy.trapz`
        kwargs : dict
            parameters of the electron distribution
        """
        gamma = np.logspace(np.log10(gamma_low), np.log10(gamma_up), 200)
        values = self.evaluate(gamma, **kwargs)
        values *= np.power(gamma, gamma_power)
        return integrator(values, gamma, axis=0)

    def integral(self, gamma_low, gamma_up, gamma_power=0):
        """integral of **this particular** electron distribution over the range 
        gamma_low, gamma_up

        Parameters
        ----------
        gamma_low : float
            lower integration limit
        gamma_up : float
            higher integration limit
        """
        gamma = np.logspace(np.log10(gamma_low), np.log10(gamma_up), 200)
        values = self.__call__(gamma)
        values *= np.power(gamma, gamma_power)
        return self.integrator(values, gamma, axis=0)

    @classmethod
    def from_normalised_density(cls, n_e_tot, **kwargs):
        r"""sets the normalisation :math:`k_e` from the total particle density 
        :math:`n_{e,\,tot}`"""
        # use gamma_min and gamma_max of the electron distribution as
        # integration limits
        if "gamma_min" in kwargs:
            gamma_min = kwargs.get("gamma_min")
        if "gamma_max" in kwargs:
            gamma_max = kwargs.get("gamma_max")
        k_e = n_e_tot / cls.general_integral(
            cls, gamma_low=gamma_min, gamma_up=gamma_max, gamma_power=0, k_e=1, **kwargs
        )
        return cls(k_e.to("cm-3"), **kwargs)

    @classmethod
    def from_normalised_energy_density(cls, u_e, **kwargs):
        r"""sets the normalisation :math:`k_e` from the total energy density 
        :math:`u_e`, Eq. 6.64 in [DermerMenon2009]_"""
        # use gamma_min and gamma_max of the electron distribution as
        # integration limits
        if "gamma_min" in kwargs:
            gamma_min = kwargs.get("gamma_min")
        if "gamma_max" in kwargs:
            gamma_max = kwargs.get("gamma_max")
        integral = cls.general_integral(
            cls, gamma_low=gamma_min, gamma_up=gamma_max, gamma_power=1, k_e=1, **kwargs
        )
        k_e = u_e / (mec2 * integral)
        return cls(k_e.to("cm-3"), **kwargs)

    @classmethod
    def from_norm_at_gamma_1(cls, norm, **kwargs):
        r"""sets :math:`k_e` such that `norm` = :math:`n_e(\gamma=1)`."""
        k_e = norm.to("cm-3") / cls.evaluate(1, 1, **kwargs)
        return cls(k_e, **kwargs)

    @classmethod
    def from_total_energy(cls, W_e, V_b, **kwargs):
        r"""sets :math:`k_e` from the total energy `W_e`, given a volume
        V_b of the emission region."""
        u_e = W_e / V_b
        return cls.from_normalised_energy_density(u_e, **kwargs)


class PowerLaw(ElectronDistribution):
    r"""Class for power-law particle spectrum. 
    When called, the particle density :math:`n_e(\gamma)` in :math:`\mathrm{cm}^{-3}` is returned.

    .. math::
        n_e(\gamma') = k_e \, \gamma'^{-p} \, H(\gamma'; \gamma'_{\rm min}, \gamma'_{\rm max}) 

    Parameters
    ----------
    k_e : :class:`~astropy.units.Quantity`
        spectral normalisation
    p : float
        spectral index, note it is positive by definition, will change sign in the function
    gamma_min : float
        minimum Lorentz factor of the electron distribution
    gamma_max : float
        maximum Lorentz factor of the electron distribution
    integrator: func
        function to be used for integration, default is :class:`~numpy.trapz`
    """

    def __init__(
        self,
        k_e=1e-13 * u.Unit("cm-3"),
        p=2.1,
        gamma_min=10,
        gamma_max=1e5,
        integrator=np.trapz,
    ):
        super().__init__(integrator)
        self.k_e = k_e
        self.p = p
        self.gamma_min = gamma_min
        self.gamma_max = gamma_max

    @property
    def parameters(self):
        return [self.k_e, self.p, self.gamma_min, self.gamma_max]

    @staticmethod
    def evaluate(gamma, k_e, p, gamma_min, gamma_max):
        return np.where(
            (gamma_min <= gamma) * (gamma <= gamma_max), k_e * gamma ** (-p), 0
        )

    def __call__(self, gamma):
        return self.evaluate(gamma, self.k_e, self.p, self.gamma_min, self.gamma_max)

    @staticmethod
    def evaluate_SSA_integrand(gamma, k_e, p, gamma_min, gamma_max):
        r"""(analytical) integrand for the synchrotron self-absorption:
        :math:`\gamma'^2 \frac{d}{d \gamma'} \left(\frac{n_e(\gamma)}{\gamma'^2}\right)`"""
        return k_e * np.where(
            (gamma_min <= gamma) * (gamma <= gamma_max),
            -(p + 2) * np.power(gamma, -p - 1),
            0,
        )

    def SSA_integrand(self, gamma):
        return self.evaluate_SSA_integrand(
            gamma, self.k_e, self.p, self.gamma_min, self.gamma_max
        )

    def __str__(self):
        return (
            f"* electron spectrum\n"
            + f" - power law\n"
            + f" - k_e: {self.k_e:.2e}\n"
            + f" - p: {self.p:.2f}\n"
            + f" - gamma_min: {self.gamma_min:.2e}\n"
            + f" - gamma_max: {self.gamma_max:.2e}\n"
        )


class BrokenPowerLaw(ElectronDistribution):
    r"""Class for broken power-law particle spectrum.
    When called, the particle density :math:`n_e(\gamma)` in :math:`\mathrm{cm}^{-3}` is returned.

    .. math::
        n_e(\gamma') = k_e \left[
        \left(\frac{\gamma'}{\gamma'_b}\right)^{-p_1} \, H(\gamma'; \gamma'_{\rm min}, \gamma'_b) +
        \left(\frac{\gamma'}{\gamma'_b}\right)^{-p_2} \, H(\gamma'; \gamma'_{b}, \gamma'_{\rm max}) 
        \right]

    Parameters
    ----------
    k_e : :class:`~astropy.units.Quantity`
        spectral normalisation
    p1 : float
        spectral index before the break (positive by definition)
    p2 : float
        spectral index after the break (positive by definition)   
    gamma_b : float
        Lorentz factor at which the change in spectral index is occurring 
    gamma_min : float
        minimum Lorentz factor of the electron distribution
    gamma_max : float
        maximum Lorentz factor of the electron distribution
    integrator: func
        function to be used for integration, default is :class:`~numpy.trapz`
    """

    def __init__(
        self,
        k_e=1e-13 * u.Unit("cm-3"),
        p1=2.0,
        p2=3.0,
        gamma_b=1e3,
        gamma_min=10,
        gamma_max=1e7,
        integrator=np.trapz,
    ):
        super().__init__(integrator)
        self.k_e = k_e
        self.p1 = p1
        self.p2 = p2
        self.gamma_b = gamma_b
        self.gamma_min = gamma_min
        self.gamma_max = gamma_max

    @property
    def parameters(self):
        return [
            self.k_e,
            self.p1,
            self.p2,
            self.gamma_b,
            self.gamma_min,
            self.gamma_max,
        ]

    @staticmethod
    def evaluate(gamma, k_e, p1, p2, gamma_b, gamma_min, gamma_max):
        index = np.where(gamma <= gamma_b, p1, p2)
        return np.where(
            (gamma_min <= gamma) * (gamma <= gamma_max),
            k_e * (gamma / gamma_b) ** (-index),
            0,
        )

    def __call__(self, gamma):
        return self.evaluate(
            gamma,
            self.k_e,
            self.p1,
            self.p2,
            self.gamma_b,
            self.gamma_min,
            self.gamma_max,
        )

    @staticmethod
    def evaluate_SSA_integrand(gamma, k_e, p1, p2, gamma_b, gamma_min, gamma_max):
        r"""(analytical) integrand for the synchrotron self-absorption:
        :math:`\gamma'^2 \frac{d}{d \gamma'} \left(\frac{n_e(\gamma)}{\gamma'^2}\right)`"""
        index = np.where(gamma <= gamma_b, p1, p2)
        return np.where(
            (gamma_min <= gamma) * (gamma <= gamma_max),
            k_e * -(index + 2) / gamma * (gamma / gamma_b) ** (-index),
            0,
        )

    def SSA_integrand(self, gamma):
        return self.evaluate_SSA_integrand(
            gamma,
            self.k_e,
            self.p1,
            self.p2,
            self.gamma_b,
            self.gamma_min,
            self.gamma_max,
        )

    def __str__(self):
        return (
            f"* electron spectrum\n"
            + f" - broken power law\n"
            + f" - k_e: {self.k_e:.2e}\n"
            + f" - p1: {self.p1:.2f}\n"
            + f" - p2: {self.p2:.2f}\n"
            + f" - gamma_b: {self.gamma_b:.2e}\n"
            + f" - gamma_min: {self.gamma_min:.2e}\n"
            + f" - gamma_max: {self.gamma_max:.2e}\n"
        )


class LogParabola(ElectronDistribution):
    r"""Class for log-parabolic particle spectrum. Built on :class:`~astropy.modeling.Fittable1DModel`.
    When called, the particle density :math:`n_e(\gamma)` in :math:`\mathrm{cm}^{-3}` is returned.

    .. math::
        n_e(\gamma') = k_e \, \left(\frac{\gamma'}{\gamma'_0}\right)^{-(p + q \log_{10}(\gamma' / \gamma'_0))}  

    Parameters
    ----------
    k_e : :class:`~astropy.units.Quantity`
        spectral normalisation
    p : float
        spectral index, note it is positive by definition, will change sign in the function
    q : float
        spectral curvature, note it is positive by definition, will change sign in the function
    gamma_0 : float
        reference Lorentz factor
    gamma_min : float
        minimum Lorentz factor of the electron distribution
    gamma_max : float
        maximum Lorentz factor of the electron distribution
    integrator: func
        function to be used for integration, default is :class:`~numpy.trapz`
    """

    def __init__(
        self,
        k_e=1e-13 * u.Unit("cm-3"),
        p=2.0,
        q=0.1,
        gamma_0=1e3,
        gamma_min=10,
        gamma_max=1e7,
        integrator=np.trapz,
    ):
        super().__init__(integrator)
        self.k_e = k_e
        self.p = p
        self.q = q
        self.gamma_0 = gamma_0
        self.gamma_min = gamma_min
        self.gamma_max = gamma_max

    @property
    def parameters(self):
        return [self.k_e, self.p, self.q, self.gamma_0, self.gamma_min, self.gamma_max]

    @staticmethod
    def evaluate(gamma, k_e, p, q, gamma_0, gamma_min, gamma_max):
        gamma_ratio = gamma / gamma_0
        index = -p - q * np.log10(gamma_ratio)
        return np.where(
            (gamma_min <= gamma) * (gamma <= gamma_max), k_e * gamma_ratio ** index, 0,
        )

    def __call__(self, gamma):
        return self.evaluate(
            gamma,
            self.k_e,
            self.p,
            self.q,
            self.gamma_0,
            self.gamma_min,
            self.gamma_max,
        )

    @staticmethod
    def evaluate_SSA_integrand(gamma, k_e, p, q, gamma_0, gamma_min, gamma_max):
        r"""(analytical) integrand for the synchrotron self-absorption:
        :math:`\gamma'^2 \frac{d}{d \gamma'} \left(\frac{n_e(\gamma)}{\gamma'^2}\right)`"""
        prefactor = -(p + 2 * q * np.log10(gamma / gamma_0) + 2) / gamma
        return prefactor * LogParabola.evaluate(
            gamma, k_e, p, q, gamma_0, gamma_min, gamma_max
        )

    def SSA_integrand(self, gamma):
        return self.evaluate_SSA_integrand(
            gamma,
            self.k_e,
            self.p,
            self.q,
            self.gamma_0,
            self.gamma_min,
            self.gamma_max,
        )

    def __str__(self):
        return (
            f"* electron spectrum\n"
            + f" - log parabola\n"
            + f" - k_e: {self.k_e:.2e}\n"
            + f" - p: {self.p:.2f}\n"
            + f" - q: {self.q:.2f}\n"
            + f" - gamma_0: {self.gamma_0:.2e}\n"
            + f" - gamma_min: {self.gamma_min:.2e}\n"
            + f" - gamma_max: {self.gamma_max:.2e}\n"
        )


class ExpCutoffPowerLaw(ElectronDistribution):
    r"""Class for power-law with an exponetial cutoff particle spectrum. 
    When called, the particle density :math:`n_e(\gamma)` in :math:`\mathrm{cm}^{-3}` is returned.

    .. math::
        n_e(\gamma') = k_e \, \gamma'^{-p} exp(-\gamma'/\gamma_c) \, H(\gamma'; \gamma'_{\rm min}, \gamma'_{\rm max}) 

    Parameters
    ----------
    k_e : :class:`~astropy.units.Quantity`
        spectral normalisation
    p : float
        spectral index, note it is positive by definition, will change sign in the function
    gamma_c : float
        cutoff Lorentz factor of the electron distribution
    gamma_min : float
        minimum Lorentz factor of the electron distribution
    gamma_max : float
        maximum Lorentz factor of the electron distribution
    integrator: func
        function to be used for integration, default is :class:`~numpy.trapz`
    """

    def __init__(
        self,
        k_e=1e-13 * u.Unit("cm-3"),
        p=2.1,
        gamma_c=1e3,
        gamma_min=10,
        gamma_max=1e5,
        integrator=np.trapz,
    ):
        super().__init__(integrator)
        self.k_e = k_e
        self.p = p
        self.gamma_c = gamma_c
        self.gamma_min = gamma_min
        self.gamma_max = gamma_max

    @property
    def parameters(self):
        return [self.k_e, self.p, self.gamma_c, self.gamma_min, self.gamma_max]

    @staticmethod
    def evaluate(gamma, k_e, p, gamma_c, gamma_min, gamma_max):
        return np.where(
            (gamma_min <= gamma) * (gamma <= gamma_max),
            k_e * gamma ** (-p) * np.exp(-gamma / gamma_c),
            0,
        )

    def __call__(self, gamma):
        return self.evaluate(
            gamma, self.k_e, self.p, self.gamma_c, self.gamma_min, self.gamma_max
        )

    @staticmethod
    def evaluate_SSA_integrand(gamma, k_e, p, gamma_c, gamma_min, gamma_max):
        r"""(analytical) integrand for the synchrotron self-absorption:
        :math:`\gamma'^2 \frac{d}{d \gamma'} \left(\frac{n_e(\gamma)}{\gamma'^2}\right)`"""
        prefactor = -(p + 2) / gamma + (-1 / gamma_c)

        return prefactor * ExpCutoffPowerLaw.evaluate(
            gamma, k_e, p, gamma_c, gamma_min, gamma_max
        )

    def SSA_integrand(self, gamma):
        return self.evaluate_SSA_integrand(
            gamma, self.k_e, self.p, self.gamma_c, self.gamma_min, self.gamma_max
        )

    def __str__(self):
        return (
            f"* electron spectrum\n"
            + f" - power law\n"
            + f" - k_e: {self.k_e:.2e}\n"
            + f" - p: {self.p:.2f}\n"
            + f" - gamma_c: {self.gamma_c:.2f}\n"
            + f" - gamma_min: {self.gamma_min:.2e}\n"
            + f" - gamma_max: {self.gamma_max:.2e}\n"
        )



# module containing the synchrotron self Compton radiative process
import numpy as np
from astropy.constants import h, c, m_e, sigma_T, G
import astropy.units as u
from .kernels import isotropic_kernel
from ..synchrotron import Synchrotron
from ..utils.math import (
    axes_reshaper,
    gamma_to_integrate,
    nu_to_integrate,
)
from ..utils.conversion import nu_to_epsilon_prime


__all__ = ["SynchrotronSelfCompton"]


class SynchrotronSelfCompton:
    """class for Synchrotron Self Compton radiation computation

    Parameters
    ----------
    blob : :class:`~agnpy.emission_region.Blob`
        emission region and electron distribution hitting the photon target
    synchrotron : :class:`~agnpy.synchrotron.Synchrotron`
        class describing the synchrotron photons target
    integrator : func
        function to be used for integration (default = `np.trapz`)
    """

    def __init__(self, blob, ssa=False, integrator=np.trapz):
        self.blob = blob
        self.ssa = ssa
        self.integrator = integrator

    @staticmethod
    def evaluate_sed_flux(
        nu,
        z,
        d_L,
        delta_D,
        B,
        R_b,
        n_e,
        *args,
        ssa=False,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) for synchrotron self-Compton,
        for a general set of model parameters. Eq. 9 in [Finke2008]_.

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D : float
            Doppler factor of the relativistic outflow
        B : :class:`~astropy.units.Quantity`
            magnetic field in the blob 
        R_b : :class:`~astropy.units.Quantity`
            size of the emitting region (spherical blob assumed)
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        ssa : bool
            whether to consider or not the self-absorption, default false
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        # synchrotron frequencies to be integrated over
        epsilon = nu_to_epsilon_prime(nu_to_integrate, z, delta_D)
        # frequencies of the final sed
        epsilon_s = nu_to_epsilon_prime(nu, z, delta_D)
        sed_synch = Synchrotron.evaluate_sed_flux(
            nu_to_integrate,
            z,
            d_L,
            delta_D,
            B,
            R_b,
            n_e,
            *args,
            ssa=ssa,
            integrator=integrator,
            gamma=gamma,
        )
        # Eq. 8 [Finke2008]_
        u_synch = (3 * np.power(d_L, 2) * sed_synch) / (
            c * np.power(R_b, 2) * np.power(delta_D, 4) * epsilon
        )
        # factor 3 / 4 accounts for averaging in a sphere
        # not included in Dermer and Finke's papers
        u_synch *= 3 / 4
        # multidimensional integration
        _gamma, _epsilon, _epsilon_s = axes_reshaper(gamma, epsilon, epsilon_s)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma, *args)
        # reshape u as epsilon
        _u_synch = np.reshape(u_synch, (1, u_synch.size, 1))
        # integrate
        kernel = isotropic_kernel(_gamma, _epsilon, _epsilon_s)
        integrand = (
            _u_synch / np.power(_epsilon, 2) * N_e / np.power(_gamma, 2) * kernel
        )
        integral_gamma = integrator(integrand, gamma, axis=0)
        integral_epsilon = integrator(integral_gamma, epsilon, axis=0)
        emissivity = 3 / 4 * c * sigma_T * np.power(epsilon_s, 2) * integral_epsilon
        prefactor = np.power(delta_D, 4) / (4 * np.pi * np.power(d_L, 2))
        return (prefactor * emissivity).to("erg cm-2 s-1")

    def sed_flux(self, nu):
        """Evaluates the SSC flux SED for a SynchrotronSelfComtpon 
        object built from a Blob."""
        return self.evaluate_sed_flux(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.B,
            self.blob.R_b,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            ssa=self.ssa,
            integrator=self.integrator,
            gamma=self.blob.gamma,
        )

    def sed_luminosity(self, nu):
        r"""Evaluates the SSC luminosity SED
        :math:`\nu L_{\nu} \, [\mathrm{erg}\,\mathrm{s}^{-1}]`
        for a a SynchrotronSelfCompton object built from a blob."""
        sphere = 4 * np.pi * np.power(self.blob.d_L, 2)
        return (sphere * self.sed_flux(nu)).to("erg s-1")

    def sed_peak_flux(self, nu):
        """provided a grid of frequencies nu, returns the peak flux of the SED
        """
        return self.sed_flux(nu).max()

    def sed_peak_nu(self, nu):
        """provided a grid of frequencies nu, returns the frequency at which the SED peaks
        """
        idx_max = self.sed_flux(nu).argmax()
        return nu[idx_max]



# module containing the gamma-gamma absorption
from pathlib import Path
import numpy as np
import astropy.units as u
from astropy.io import fits
from astropy.constants import c, G, h, m_e, M_sun, sigma_T
from scipy.interpolate import interp2d
from ..utils.math import (
    axes_reshaper,
    log,
    mu_to_integrate,
    phi_to_integrate,
    min_rel_distance,
)
from ..utils.geometry import (
    cos_psi,
    x_re_shell,
    mu_star_shell,
    x_re_ring,
    x_re_ring_mu_s,
    phi_mu_re_shell,
    phi_mu_re_ring,
    x_re_shell_mu_s,
)
from ..utils.conversion import nu_to_epsilon_prime, to_R_g_units
from ..targets import PointSourceBehindJet, SSDisk, SphericalShellBLR, RingDustTorus
from ..emission_regions import Blob
from ..synchrotron import nu_synch_peak, Synchrotron


__all__ = ["sigma", "Absorption", "ebl_files_dict", "EBL"]

agnpy_dir = Path(__file__).parent.parent
ebl_files_dict = {
    "franceschini": f"{agnpy_dir}/data/ebl_models/ebl_franceschini08.fits.gz",
    "dominguez": f"{agnpy_dir}/data/ebl_models/ebl_dominguez11.fits.gz",
    "finke": f"{agnpy_dir}/data/ebl_models/ebl_finke10.fits.gz",
}


def sigma(s):
    """photon-photon pair production cross section, Eq. 17 of [Dermer2009]"""
    beta_cm = np.sqrt(1 - 1 / s)
    prefactor = 3 / 16 * sigma_T * (1 - np.power(beta_cm, 2))
    term1 = (3 - np.power(beta_cm, 4)) * log((1 + beta_cm) / (1 - beta_cm))
    term2 = -2 * beta_cm * (2 - np.power(beta_cm, 2))
    values = prefactor * (term1 + term2)
    values[s < 1] = 0
    return values


class Absorption:
    """class to compute the absorption due to gamma-gamma pair production

    Parameters
    ----------
    blob : :class:`~agnpy.emission_regions.Blob`
        emission region and electron distribution hitting the photon target
    target : :class:`~agnpy.targets` or class:`~agnpy.emission_regions.Blob`
        class describing the target photon field
    r : :class:`~astropy.units.Quantity`
        distance of the blob from the Black Hole (i.e. from the target photons)
        the distance is irrelevant in the case of absorption
    """

    def __init__(self, target, r=None, z=0, mu_s=1):
        self.target = target
        self.r = r
        self.z = z
        self.mu_s = mu_s
        self.set_mu()
        self.set_phi()
        self.set_l()
        # r can be only ignored for absorption on synchrotron radiation
        if r is None and not isinstance(self.target, Blob):
            raise ValueError(
                "No distance provided for absorption on "
                + str(target.__class__)
                + ", this can be only done for Blob class"
            )

    def set_mu(self, mu_size=100):
        self.mu_size = mu_size
        self.mu = np.linspace(-1, 1, self.mu_size)

    def set_phi(self, phi_size=50):
        self.phi_size = phi_size
        self.phi = np.linspace(0, 2 * np.pi, self.phi_size)

    def set_l(self, l_size=50):
        self.l_size = l_size

    @staticmethod
    def evaluate_tau_ps_behind_blob(nu, z, mu_s, epsilon_0, L_0, r):
        r"""Evaluates the absorption produced by the photon field of a point 
        source of photons behind the blob, for a general set of model parameters

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the opacity 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        epsilon_0 : float
            dimensionless energy (in electron rest mass energy units) of the
            target photon field
        L_0 : :class:`~astropy.units.Quantity`
            luminosity [erg cm-3] of the point source behind the jet
        r : :class:`~astropy.units.Quantity`
            distance between the point source and the blob
        
        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the tau values corresponding to each frequency
        """
        epsilon_1 = nu_to_epsilon_prime(nu, z)
        s = epsilon_0 * epsilon_1 * (1 - mu_s) / 2
        integral = (1 - mu_s) * sigma(s) / r
        prefactor = L_0 / (4 * np.pi * epsilon_0 * m_e * c ** 3)
        return (prefactor * integral).to_value("")

    def tau_ps_behind_blob(self, nu):
        """Evaluates the absorption produced by the photon field of a point 
        source of photons behind the blob"""
        return self.evaluate_tau_ps_behind_blob(
            nu, self.z, self.mu_s, self.target.epsilon_0, self.target.L_0, self.r
        )

    @staticmethod
    def evaluate_tau_ps_behind_blob_mu_s(nu, z, mu_s, epsilon_0, L_0, r, u_size=100):
        r"""Evaluates the absorption produced by the photon field of a point 
        source of photons behind the blob, for a general set of model parameters

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the opacity 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        epsilon_0 : float
            dimensionless energy (in electron rest mass energy units) of the
            target photon field
        L_0 : :class:`~astropy.units.Quantity`
            luminosity [erg cm-3] of the point source behind the jet
        r : :class:`~astropy.units.Quantity`
            distance between the point source and the blob
        u_size : int
            size of the array of distances from the photon origin to integrate over
        
        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the tau values corresponding to each frequency
        """
        epsilon_1 = nu_to_epsilon_prime(nu, z)

        uu = np.logspace(-5, 5, u_size) * r
        _u, _epsilon_1 = axes_reshaper(uu, epsilon_1)
        # distance between soft photon and gamma ray
        x = np.sqrt(r * r + _u * _u + 2 * _u * r * mu_s)

        # cos angle of the soft photon to the z axis
        _mu = (r + _u * mu_s) / x
        phi = 0  # both gamma ray and soft photon move in XZ plane
        _cos_psi = cos_psi(mu_s, _mu, phi)
        s = _epsilon_1 * epsilon_0 * (1 - _cos_psi) / 2

        integrand = (1 - _cos_psi) / x ** 2 * sigma(s)
        # integrate
        integral = np.trapz(integrand, uu, axis=0)
        prefactor = L_0 / (4 * np.pi * epsilon_0 * m_e * c ** 3)
        return (prefactor * integral).to_value("")

    def tau_ps_behind_blob_mu_s(self, nu):
        """Evaluates the absorption produced by the photon field of a point 
        source of photons behind the blob"""
        return self.evaluate_tau_ps_behind_blob_mu_s(
            nu, self.z, self.mu_s, self.target.epsilon_0, self.target.L_0, self.r
        )

    @staticmethod
    def evaluate_tau_ss_disk(
        nu,
        z,
        mu_s,
        M_BH,
        L_disk,
        eta,
        R_in,
        R_out,
        r,
        R_tilde_size=100,
        l_tilde_size=50,
        phi=phi_to_integrate,
    ):
        """Evaluates the gamma-gamma absorption produced by the photon field of 
        a Shakura-Sunyaev accretion disk

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the opacity 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        M_BH : :class:`~astropy.units.Quantity`
            Black Hole mass    
        L_disk : :class:`~astropy.units.Quantity`
            luminosity of the disk 
        eta : float
            accretion efficiency
        R_in : :class:`~astropy.units.Quantity` 
            inner disk radius
        R_out : :class:`~astropy.units.Quantity` 
            inner disk radius
        R_tilde_size : int
            size of the array of disk coordinates to integrate over
        r : :class:`~astropy.units.Quantity`
            distance between the point source and the blob
        l_tilde_size : int
            size of the array of distances from the BH to integrate over
        phi : :class:`~numpy.ndarray`
            array of azimuth angles to integrate over
        
        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the tau values corresponding to each frequency
        """
        # conversions
        R_g = (G * M_BH / c ** 2).to("cm")
        r_tilde = to_R_g_units(r, M_BH)
        R_in_tilde = to_R_g_units(R_in, M_BH)
        R_out_tilde = to_R_g_units(R_out, M_BH)
        # multidimensional integration
        R_tilde = np.linspace(R_in_tilde, R_out_tilde, R_tilde_size)
        l_tilde = np.logspace(0, 5, l_tilde_size) * r_tilde
        epsilon_1 = nu_to_epsilon_prime(nu, z)
        _R_tilde, _phi, _l_tilde, _epsilon_1 = axes_reshaper(
            R_tilde, phi, l_tilde, epsilon_1
        )
        _epsilon = SSDisk.evaluate_epsilon(L_disk, M_BH, eta, _R_tilde)
        _phi_disk = 1 - (R_in_tilde / _R_tilde) ** (1 / 2)
        _mu = (1 + (_R_tilde ** 2 / _l_tilde ** 2)) ** (-1 / 2)
        _cos_psi = cos_psi(mu_s, _mu, _phi)
        s = _epsilon * _epsilon_1 * (1 - _cos_psi) / 2
        integrand = (
            1
            / _l_tilde ** 2
            / _R_tilde ** 2
            / (1 + (_R_tilde ** 2 / _l_tilde ** 2)) ** (3 / 2)
            * _phi_disk
            / _epsilon
            * sigma(s)
            * (1 - _cos_psi)
        )
        integral_R_tilde = np.trapz(integrand, R_tilde, axis=0)
        integral_phi = np.trapz(integral_R_tilde, phi, axis=0)
        integral = np.trapz(integral_phi, l_tilde, axis=0)
        prefactor = 3 * L_disk / ((4 * np.pi) ** 2 * eta * m_e * c ** 3 * R_g)
        return (prefactor * integral).to_value("")

    def tau_ss_disk(self, nu):
        """Evaluates the gamma-gamma absorption produced by the photon field of 
        a Shakura-Sunyaev accretion disk"""
        return self.evaluate_tau_ss_disk(
            nu,
            self.z,
            self.mu_s,
            self.target.M_BH,
            self.target.L_disk,
            self.target.eta,
            self.target.R_in,
            self.target.R_out,
            self.r,
            R_tilde_size=100,
            l_tilde_size=self.l_size,
            phi=self.phi,
        )

    @staticmethod
    def evaluate_tau_blr(
        nu,
        z,
        mu_s,
        L_disk,
        xi_line,
        epsilon_line,
        R_line,
        r,
        l_size=50,
        mu=mu_to_integrate,
        phi=phi_to_integrate,
    ):
        """Evaluates the gamma-gamma absorption produced by a spherical shell 
        BLR for a general set of model parameters

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the tau 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        L_disk : :class:`~astropy.units.Quantity`
            Luminosity of the disk whose radiation is being reprocessed by the BLR
        xi_line : float
            fraction of the disk radiation reprocessed by the BLR
        epsilon_line : string
            dimensionless energy of the emitted line
        R_line : :class:`~astropy.units.Quantity`
            radius of the BLR spherical shell
        r : :class:`~astropy.units.Quantity`
            distance between the Broad Line Region and the blob
        l_size : int
            size of the array of distances from the BH to integrate over
        mu, phi : :class:`~numpy.ndarray`
            arrays of cosine of zenith and azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the tau values corresponding to each frequency
        """
        # conversions
        epsilon_1 = nu_to_epsilon_prime(nu, z)
        # multidimensional integration
        l = np.logspace(0, 5, l_size) * r

        # check if any point is too close to R_line, the function works only for mu=1, so
        # we can check directly if R_line is within 'l' array
        idx = np.isclose(l, R_line, rtol=min_rel_distance)
        l[idx] += min_rel_distance * R_line

        _mu, _phi, _l, _epsilon_1 = axes_reshaper(mu, phi, l, epsilon_1)
        x = x_re_shell(_mu, R_line, _l)
        _mu_star = mu_star_shell(_mu, R_line, _l)
        _cos_psi = cos_psi(mu_s, _mu_star, _phi)
        s = _epsilon_1 * epsilon_line * (1 - _cos_psi) / 2
        integrand = (1 - _cos_psi) / x ** 2 * sigma(s)
        # integrate
        integral_mu = np.trapz(integrand, mu, axis=0)
        integral_phi = np.trapz(integral_mu, phi, axis=0)
        integral = np.trapz(integral_phi, l, axis=0)
        prefactor = (L_disk * xi_line) / (
            (4 * np.pi) ** 2 * epsilon_line * m_e * c ** 3
        )
        return (prefactor * integral).to_value("")

    @staticmethod
    def evaluate_tau_blr_mu_s(
        nu,
        z,
        mu_s,
        L_disk,
        xi_line,
        epsilon_line,
        R_line,
        r,
        u_size=100,
        mu=mu_to_integrate,
        phi=phi_to_integrate,
    ):
        """Evaluates the gamma-gamma absorption produced by a spherical shell 
        BLR for a general set of model parameters and arbitrary mu_s

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the tau 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        L_disk : :class:`~astropy.units.Quantity`
            Luminosity of the disk whose radiation is being reprocessed by the BLR
        xi_line : float
            fraction of the disk radiation reprocessed by the BLR
        epsilon_line : string
            dimensionless energy of the emitted line
        R_line : :class:`~astropy.units.Quantity`
            radius of the BLR spherical shell
        r : :class:`~astropy.units.Quantity`
            distance between the Broad Line Region and the blob
        l_size : int
            size of the array of distances from the BH to integrate over
        mu, phi : :class:`~numpy.ndarray`
            arrays of cosine of zenith and azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the tau values corresponding to each frequency
        """
        # conversions
        epsilon_1 = nu_to_epsilon_prime(nu, z)
        # multidimensional integration
        # here uu is the distance that the photon traversed
        uu = np.logspace(-5, 5, u_size) * r

        # check if for any uu value the position of the photon is too close to the BLR
        x_cross = np.sqrt(r ** 2 + uu ** 2 + 2 * uu * r * mu_s)
        idx = np.isclose(x_cross, R_line, rtol=min_rel_distance)
        if idx.any():
            uu[idx] += min_rel_distance * R_line
            # it might happen that some of the points get more shifted then the next one,
            # possibly making integration messy, so we sort the points
            uu = np.sort(uu)

        _mu_re, _phi_re, _u, _epsilon_1 = axes_reshaper(mu, phi, uu, epsilon_1)

        # distance between soft photon and gamma ray
        x = x_re_shell_mu_s(R_line, r, _phi_re, _mu_re, _u, mu_s)

        # convert the phi and mu angles of the position in the sphere into the actual phi and mu angles
        # the actual phi and mu angles of the soft photon catching up with the gamma ray
        _phi, _mu_star = phi_mu_re_shell(R_line, r, _phi_re, _mu_re, _u, mu_s)

        # angle between the soft photon and gamma ray
        _cos_psi = cos_psi(mu_s, _mu_star, _phi)
        s = _epsilon_1 * epsilon_line * (1 - _cos_psi) / 2
        integrand = (1 - _cos_psi) / x ** 2 * sigma(s)
        # integrate
        integral_mu = np.trapz(integrand, mu, axis=0)
        integral_phi = np.trapz(integral_mu, phi, axis=0)
        integral = np.trapz(integral_phi, uu, axis=0)
        prefactor = (L_disk * xi_line) / (
            (4 * np.pi) ** 2 * epsilon_line * m_e * c ** 3
        )
        return (prefactor * integral).to_value("")

    def tau_blr(self, nu):
        """Evaluates the gamma-gamma absorption produced by a spherical shell 
        BLR for a general set of model parameters
        """
        return self.evaluate_tau_blr(
            nu,
            self.z,
            self.mu_s,
            self.target.L_disk,
            self.target.xi_line,
            self.target.epsilon_line,
            self.target.R_line,
            self.r,
            l_size=self.l_size,
            mu=self.mu,
            phi=self.phi,
        )

    def tau_blr_mu_s(self, nu):
        """Evaluates the gamma-gamma absorption produced by a spherical shell 
        BLR for a general set of model parameters and arbitrary mu_s
        """
        return self.evaluate_tau_blr_mu_s(
            nu,
            self.z,
            self.mu_s,
            self.target.L_disk,
            self.target.xi_line,
            self.target.epsilon_line,
            self.target.R_line,
            self.r,
            u_size=2 * self.l_size,
            mu=self.mu,
            phi=self.phi,
        )

    @staticmethod
    def evaluate_tau_dt(
        nu,
        z,
        mu_s,
        L_disk,
        xi_dt,
        epsilon_dt,
        R_dt,
        r,
        l_size=50,
        phi=phi_to_integrate,
    ):
        r"""Evaluates the gamma-gamma absorption produced by a ring dust torus

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        L_disk : :class:`~astropy.units.Quantity`
            Luminosity of the disk whose radiation is being reprocessed by the BLR
        xi_dt : float
            fraction of the disk radiation reprocessed by the disk
        epsilon_dt : string
            peak (dimensionless) energy of the black body radiated by the torus 
        R_dt : :class:`~astropy.units.Quantity`
            radius of the ting-like torus
        r : :class:`~astropy.units.Quantity`
            distance between the dust torus and the blob
        l_size : int
            size of the array of distances from the BH to integrate over
        phi : :class:`~numpy.ndarray`
            arrays of azimuth angles to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        epsilon_1 = nu_to_epsilon_prime(nu, z)
        # multidimensional integration
        l = np.logspace(0, 5, l_size) * r
        _phi, _l, _epsilon_1 = axes_reshaper(phi, l, epsilon_1)
        x = x_re_ring(R_dt, _l)
        _mu = _l / x
        _cos_psi = cos_psi(mu_s, _mu, _phi)
        s = _epsilon_1 * epsilon_dt * (1 - _cos_psi) / 2
        integrand = (1 - _cos_psi) / x ** 2 * sigma(s)
        # integrate
        integral_phi = np.trapz(integrand, phi, axis=0)
        integral = np.trapz(integral_phi, l, axis=0)
        prefactor = (L_disk * xi_dt) / (8 * np.pi ** 2 * epsilon_dt * m_e * c ** 3)
        return (prefactor * integral).to_value("")

    @staticmethod
    def evaluate_tau_dt_mu_s(
        nu,
        z,
        mu_s,
        L_disk,
        xi_dt,
        epsilon_dt,
        R_dt,
        r,
        u_size=100,
        phi_re=phi_to_integrate,
    ):
        r"""Evaluates the gamma-gamma absorption produced by a ring dust torus
        for the case of photon moving at an angle to the jet 

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        mu_s : float
            cosine of the angle between the blob motion and the jet axis
        L_disk : :class:`~astropy.units.Quantity`
            Luminosity of the disk whose radiation is being reprocessed by the BLR
        xi_dt : float
            fraction of the disk radiation reprocessed by the disk
        epsilon_dt : string
            peak (dimensionless) energy of the black body radiated by the torus 
        R_dt : :class:`~astropy.units.Quantity`
            radius of the ting-like torus
        r : :class:`~astropy.units.Quantity`
            distance between the dust torus and the blob
        u_size : int
            size of the array of distances from the photon origin to integrate over
        phi_re : :class:`~numpy.ndarray`
            arrays of azimuth angles of the dust torus to integrate over

        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        epsilon_1 = nu_to_epsilon_prime(nu, z)
        # multidimensional integration
        # here uu is the distance that the photon traversed
        uu = np.logspace(-5, 5, u_size) * r
        _phi_re, _u, _epsilon_1 = axes_reshaper(phi_re, uu, epsilon_1)
        # distance between soft photon and gamma ray
        x = x_re_ring_mu_s(R_dt, r, _phi_re, _u, mu_s)
        # convert the phi angles of the ring into the actual phi angles
        # of the soft photon catching up with the gamma ray
        _phi, _mu = phi_mu_re_ring(R_dt, r, _phi_re, _u, mu_s)
        _cos_psi = cos_psi(mu_s, _mu, _phi)
        s = _epsilon_1 * epsilon_dt * (1 - _cos_psi) / 2
        integrand = (1 - _cos_psi) / x ** 2 * sigma(s)
        # integrate
        integral_phi = np.trapz(integrand, phi_re, axis=0)
        integral = np.trapz(integral_phi, uu, axis=0)
        prefactor = (L_disk * xi_dt) / (8 * np.pi ** 2 * epsilon_dt * m_e * c ** 3)
        return (prefactor * integral).to_value("")

    def tau_dt(self, nu):
        """evaluates the gamma-gamma absorption produced by a ring dust torus"""
        return self.evaluate_tau_dt(
            nu,
            self.z,
            self.mu_s,
            self.target.L_disk,
            self.target.xi_dt,
            self.target.epsilon_dt,
            self.target.R_dt,
            self.r,
            l_size=self.l_size,
            phi=self.phi,
        )

    def tau_dt_mu_s(self, nu):
        """evaluates the gamma-gamma absorption produced by a ring dust torus"""
        return self.evaluate_tau_dt_mu_s(
            nu,
            self.z,
            self.mu_s,
            self.target.L_disk,
            self.target.xi_dt,
            self.target.epsilon_dt,
            self.target.R_dt,
            self.r,
            u_size=2 * self.l_size,
            phi_re=self.phi,
        )

    def tau_on_synchrotron(self, blob, nu, nu_s_size=200, delta_margin_low=1.0e-2):
        r"""Optical depth for absorption of gamma rays in synchrotron radiation of the blob.
        It assumes the same radiation field as the SSC class.

        Parameters
        ----------
        blob : :class:`~agnpy.emission_regions.Blob`
            emission region and electron distribution hitting the photon target
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the opacity
            **note** these are observed frequencies (observer frame)
        nu_s_size : int
            size of the array over the synchrotron frequencies
        delta_margin_low : float
            extension of the integration range of the synchrotron radiation beyond
            the delta approximation, default = 0.01, but lower value might be needed
            if the calculations are performed up to very high energies
        """
        # energy of the gamma rays in blob frame
        epsilon1 = nu_to_epsilon_prime(nu, blob.z, blob.delta_D)

        # first derive the ranges of the synchrotron spectrum using delta approximation
        # add margin on both sides to allow for the energy distribution
        nu_s_min = nu_synch_peak(blob.B, blob.gamma_min) * delta_margin_low
        nu_s_max = nu_synch_peak(blob.B, blob.gamma_max) * 1.0e2

        # frequencies in the blob frame
        nu_s = (
            np.logspace(
                np.log10(nu_s_min.to_value("Hz")),
                np.log10(nu_s_max.to_value("Hz")),
                nu_s_size,
            )
            * u.Hz
        )

        # and in observers frame
        nu_s_obs = nu_s * blob.delta_D / (1 + blob.z)
        # energy of the synchrotron photons in blob frame
        epsilon = nu_to_epsilon_prime(nu_s_obs, blob.z, blob.delta_D)

        synch = Synchrotron(blob, ssa=True)
        sed_synch = synch.sed_flux(nu_s_obs)

        # Eq. 8 [Finke2008]_ divided by extra epsilon mc^2
        n_synch = (
            (3 * np.power(blob.d_L, 2) * sed_synch)
            / (
                c
                * np.power(blob.R_b, 2)
                * np.power(blob.delta_D, 4)
                * epsilon ** 2
                * m_e
                * c ** 2
            )
        ).to("cm-3")

        # factor 3 / 4 accounts for averaging in a sphere
        # not included in Dermer and Finke's papers
        n_synch *= 3 / 4

        _epsilon, _epsilon1 = axes_reshaper(epsilon, epsilon1)
        _s = _epsilon * _epsilon1 / 2
        _n_synch = n_synch[..., np.newaxis]

        return (2 * blob.R_b * np.trapz(_n_synch * sigma(_s), epsilon, axis=0)).to("")

    def tau(self, nu):
        """optical depth

        .. math::
            \\tau_{\\gamma \\gamma}(\\nu)

        Parameters
        ----------
        nu : `~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the opacity, **note** these are
            observed frequencies (observer frame).
        """

        if isinstance(self.target, Blob):
            return self.tau_on_synchrotron(self.target, nu)

        if self.mu_s == 1:  # default value
            if isinstance(self.target, PointSourceBehindJet):
                return self.tau_ps_behind_blob(nu)  # this is always 0 for mu_s=1!
            if isinstance(self.target, SSDisk):
                return self.tau_ss_disk(nu)
            if isinstance(self.target, SphericalShellBLR):
                return self.tau_blr(nu)
            if isinstance(self.target, RingDustTorus):
                return self.tau_dt(nu)
        else:
            if isinstance(self.target, PointSourceBehindJet):
                return self.tau_ps_behind_blob_mu_s(nu)
            # those are yet to be implemented
            # if isinstance(self.target, SSDisk):
            #    return self.tau_ss_disk(nu)
            if isinstance(self.target, SphericalShellBLR):
                return self.tau_blr_mu_s(nu)
            if isinstance(self.target, RingDustTorus):
                return self.tau_dt_mu_s(nu)

    def absorption(self, nu):
        """This function returns the attenuation of the emission assuming that
        the optical depth tau is computed from the production place to the observer.
        """
        return np.exp(-self.tau(nu))

    def absorption_homogeneous(self, nu):
        """This function returns the attenuation of the emission assuming that
        the emission is produced homogenously inside absorbing material.
        The calculations is only accurate for a slab of absorbing material with the
        total optical depth tau, but the same formula is often used also e.g.
        in the context of absorption of gamma-ray emission by synchrotron radiation in blobs
        See e.g. section 2.5.1. of Finke et al. 2008.
        """
        t = self.tau(nu)
        return (1 - np.exp(-t)) / t


class EBL:
    """Class representing for the Extragalactic Background Light absorption. 
    Tabulated values of absorption as a function of redshift and energy according
    to the models of [Franceschini2008]_, [Finke2010]_, [Dominguez2011]_ are available
    in `data/ebl_models`. 
    They are interpolated by `agnpy` and can be later evaluated for a given redshift 
    and range of frequencies.

    Parameters
    ----------
    model : ["franceschini", "dominguez", "finke"]
        choose the reference for the EBL model
    """

    def __init__(self, model="franceschini"):
        if model not in ["franceschini", "dominguez", "finke"]:
            raise ValueError("No EBL model for the reference you specified")
        else:
            self.model_file = ebl_files_dict[model]
        # load the absorption table
        self.load_absorption_table()
        self.interpolate_absorption_table()

    def load_absorption_table(self):
        """load the reference values from the table file to be interpolated later"""
        f = fits.open(self.model_file)
        self.energy_ref = (
            np.sqrt(f["ENERGIES"].data["ENERG_LO"] * f["ENERGIES"].data["ENERG_HI"])
            * u.eV
        )
        # Franceschini file has two columns repeated, eliminate them
        self.z_ref = np.unique(f["SPECTRA"].data["PARAMVAL"])
        self.values_ref = np.unique(f["SPECTRA"].data["INTPSPEC"], axis=0)

    def interpolate_absorption_table(self, kind="linear"):
        """interpolate the reference values, choose the kind of interpolation"""
        log10_energy_ref = np.log10(self.energy_ref.to_value("eV"))
        self.interpolated_model = interp2d(
            log10_energy_ref, self.z_ref, self.values_ref, kind=kind
        )

    def absorption(self, z, nu):
        energy = nu.to_value("eV", equivalencies=u.spectral())
        log10_energy = np.log10(energy)
        return self.interpolated_model(log10_energy, z)



# test on synchrotron module
import pytest
import numpy as np
import astropy.units as u
from astropy.constants import m_e, c, h
from astropy.coordinates import Distance
from pathlib import Path
from agnpy.emission_regions import Blob
from agnpy.synchrotron import Synchrotron, nu_synch_peak, epsilon_B
from agnpy.utils.math import trapz_loglog
from .utils import (
    make_comparison_plot,
    extract_columns_sample_file,
    check_deviation,
)


agnpy_dir = Path(__file__).parent.parent
# where to read sampled files
data_dir = agnpy_dir / "data"
# where to save figures
figures_dir = agnpy_dir.parent / "crosschecks/figures/synchrotron"
figures_dir.mkdir(parents=True, exist_ok=True)


# variables with _test are global and meant to be used in all tests
# here as a default we use the same parameters of Figure 7.4 in Dermer Menon 2009
spectrum_norm_test = 1e48 * u.Unit("erg")
p_test = 2.8
q_test = 0.2
gamma_0 = 1e3
gamma_min_test = 1e2
gamma_max_test = 1e5
pwl_dict_test = {
    "type": "PowerLaw",
    "parameters": {
        "p": p_test,
        "gamma_min": gamma_min_test,
        "gamma_max": gamma_max_test,
    },
}
# dictionary test for log parabola
lp_dict_test = {
    "type": "LogParabola",
    "parameters": {
        "p": p_test,
        "q": q_test,
        "gamma_0": gamma_0,
        "gamma_min": gamma_min_test,
        "gamma_max": gamma_max_test,
    },
}
# blob parameters
R_b_test = 1e16 * u.cm
B_test = 1 * u.G
z_test = Distance(1e27, unit=u.cm).z
delta_D_test = 10
Gamma_test = 10
pwl_blob_test = Blob(
    R_b_test,
    z_test,
    delta_D_test,
    Gamma_test,
    B_test,
    spectrum_norm_test,
    pwl_dict_test,
)
lp_blob_test = Blob(
    R_b_test,
    z_test,
    delta_D_test,
    Gamma_test,
    B_test,
    spectrum_norm_test,
    lp_dict_test,
)


class TestSynchrotron:
    """class grouping all tests related to the Synchrotron class"""

    @pytest.mark.parametrize("gamma_max, nu_range_max", [("1e5", 1e18), ("1e7", 1e22)])
    def test_synch_reference_sed(self, gamma_max, nu_range_max):
        """test agnpy synchrotron SED against the ones in Figure 7.4 of Dermer 
        Menon 2009"""
        # reference SED
        nu_ref, sed_ref = extract_columns_sample_file(
            f"{data_dir}/reference_seds/dermer_menon_2009/figure_7_4/synchrotron_gamma_max_{gamma_max}.txt",
            "Hz",
            "erg cm-2 s-1",
        )
        # agnpy
        # change the gamma_max in the blob
        pwl_dict_test["parameters"]["gamma_max"] = float(gamma_max)
        pwl_blob_test.set_spectrum(spectrum_norm_test, pwl_dict_test, "integral")
        # recompute the SED at the same ordinates of the reference figures
        synch = Synchrotron(pwl_blob_test)
        sed_agnpy = synch.sed_flux(nu_ref)
        # sed comparison plot
        nu_range = [1e10, nu_range_max] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "Figure 7.4, Dermer and Menon (2009)",
            "Synchrotron, " + r"$\gamma_{max} = $" + gamma_max,
            f"{figures_dir}/synch_comparison_gamma_max_{gamma_max}_figure_7_4_dermer_menon_2009.png",
            "sed",
            y_range=[1e-13, 1e-9],
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 25% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.25, nu_range)

    @pytest.mark.parametrize(
        "file_ref , spectrum_type, spectrum_parameters, figure_title, figure_path",
        [
            (
                f"{data_dir}/reference_seds/jetset/data/synch_ssa_pwl_jetset_1.1.2.txt",
                "PowerLaw",
                {"p": 2, "gamma_min": 2, "gamma_max": 1e6},
                "Self-Absorbed Synchrotron, power-law electron distribution",
                f"{figures_dir}/ssa_pwl_comparison_jetset_1.1.2.png",
            ),
            (
                f"{data_dir}/reference_seds/jetset/data/synch_ssa_bpwl_jetset_1.1.2.txt",
                "BrokenPowerLaw",
                {"p1": 2, "p2": 3, "gamma_b": 1e4, "gamma_min": 2, "gamma_max": 1e6},
                "Self-Absorbed Synchrotron, broken power-law electron distribution",
                f"{figures_dir}/ssa_bpwl_comparison_jetset_1.1.2.png",
            ),
            (
                f"{data_dir}/reference_seds/jetset/data/synch_ssa_lp_jetset_1.1.2.txt",
                "LogParabola",
                {"p": 2, "q": 0.4, "gamma_0": 1e4, "gamma_min": 2, "gamma_max": 1e6},
                "Self-Absorbed Synchrotron, log-parabola electron distribution",
                f"{figures_dir}/ssa_lp_comparison_jetset_1.1.2.png",
            ),
        ],
    )
    def test_ssa_reference_sed(
        self, file_ref, spectrum_type, spectrum_parameters, figure_title, figure_path,
    ):
        """test SSA SED generated by a given electron distribution against the 
        ones generated with jetset version 1.1.2, via jetset_ssa_sed.py script"""
        # reference SED
        nu_ref, sed_ref = extract_columns_sample_file(file_ref, "Hz", "erg cm-2 s-1")
        # same parameters used to produce the jetset SED
        spectrum_norm = 1e2 * u.Unit("cm-3")
        spectrum_dict = {"type": spectrum_type, "parameters": spectrum_parameters}
        blob = Blob(
            R_b=5e15 * u.cm,
            z=0.1,
            delta_D=10,
            Gamma=10,
            B=0.1 * u.G,
            spectrum_norm=spectrum_norm,
            spectrum_dict=spectrum_dict,
        )
        # recompute the SED at the same ordinates where the figure was sampled
        ssa = Synchrotron(blob, ssa=True)
        sed_agnpy = ssa.sed_flux(nu_ref)
        # sed comparison plot, we will check between 10^(11) and 10^(19) Hz
        nu_range = [1e11, 1e19] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "jetset 1.1.2",
            figure_title,
            figure_path,
            "sed",
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 5% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.05, nu_range)

    def test_synch_delta_sed(self):
        """check that in a given frequency range the full synchrotron SED coincides
        with the delta function approximation"""
        nu = np.logspace(10, 20) * u.Hz
        synch = Synchrotron(lp_blob_test)
        sed_full = synch.sed_flux(nu)
        sed_delta = synch.sed_flux_delta_approx(nu)
        # range of comparison
        nu_range = [1e12, 1e17] * u.Hz
        make_comparison_plot(
            nu,
            sed_delta,
            sed_full,
            "delta function approximation",
            "full integration",
            "Synchrotron",
            f"{figures_dir}/synch_comparison_delta_aprproximation.png",
            "sed",
            [1e-16, 1e-8],
            nu_range.to_value("Hz"),
        )
        # requires that the delta approximation SED points deviate less than 10%
        assert check_deviation(nu, sed_delta, sed_full, 0.1, nu_range)

    def test_sed_integration_methods(self):
        """test different integration methods agains each other
        simpole trapezoidal rule vs trapezoidal rule in log-log space
        """
        nu = np.logspace(8, 22) * u.Hz
        synch_trapz = Synchrotron(pwl_blob_test, integrator=np.trapz)
        synch_trapz_loglog = Synchrotron(pwl_blob_test, integrator=trapz_loglog)
        sed_synch_trapz = synch_trapz.sed_flux(nu)
        sed_synch_trapz_loglog = synch_trapz_loglog.sed_flux(nu)
        make_comparison_plot(
            nu,
            sed_synch_trapz_loglog,
            sed_synch_trapz,
            "trapezoidal log-log integration",
            "trapezoidal integration",
            "Synchrotron",
            f"{figures_dir}/synch_comparison_integration_methods.png",
            "sed",
        )
        # requires that the SED points deviate less than 1%
        assert check_deviation(nu, sed_synch_trapz_loglog, sed_synch_trapz, 0.01)

    def test_nu_synch_peak(self):
        gamma = 100
        nu_synch = nu_synch_peak(B_test, gamma).to_value("Hz")
        assert np.isclose(nu_synch, 27992489872.33304, atol=0)

    def test_epsilon_B(self):
        assert np.isclose(epsilon_B(B_test), 2.2655188038060715e-14, atol=0)
